# -*- encoding: utf-8 -*-
# ruff: noqa: F401
"""The Nendo AI Audio Tool Suite."""

from importlib import metadata

from .config import NendoConfig
from .library import SqlAlchemyNendoLibrary
from .main import Nendo
from .schema import (
    NendoAnalysisPlugin,
    NendoBucketNotFoundError,
    NendoCollection,
    NendoCollectionNotFoundError,
    NendoCollectionSlim,
    NendoEffectPlugin,
    NendoError,
    NendoFileFormatError,
    NendoGeneratePlugin,
    NendoIdError,
    NendoLibraryError,
    NendoLibraryIdError,
    NendoLibraryPlugin,
    NendoPlugin,
    NendoPluginConfigError,
    NendoPluginData,
    NendoPluginDataNotFoundError,
    NendoPluginError,
    NendoPluginLoadingError,
    NendoPluginRuntimeError,
    NendoRelationshipNotFoundError,
    NendoStorage,
    NendoStorageError,
    NendoStorageLocalFS,
    NendoTrack,
    NendoTrackNotFoundError,
    NendoTrackSlim,
    NendoUser,
    NendoUserNotFoundError,
    ResourceLocation,
)

meta = metadata.metadata(__package__ or __name__)

__version__ = meta["Version"]
__author__ = meta["Author"]
__email__ = meta["Author-email"]
__description__ = meta["Description"]
__url__ = meta["Project-URL"]



# -*- encoding: utf-8 -*-
# ruff: noqa: A003
"""Module containing all SQLAlchemy ORM models for the nendo core schema.

Used by the SQLAlchemy implementation of the nendo library.
"""

from __future__ import annotations

import json
import uuid
from datetime import date, datetime

import numpy as np
from sqlalchemy import Column, DateTime, ForeignKey, Integer, MetaData, String
from sqlalchemy.dialects.postgresql import ENUM, JSON, UUID
from sqlalchemy.orm import declarative_base, relationship
from sqlalchemy.sql import func
from sqlalchemy.sql.sqltypes import Text
from sqlalchemy.types import TypeDecorator
from sqlalchemy_json import NestedMutableDict, NestedMutableList, mutable_json_type

from nendo import schema

Base = declarative_base(metadata=MetaData())


def convert(obj):
    if isinstance(obj, uuid.UUID):
        return str(obj)
    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    if isinstance(obj, set):
        return convert(list(obj))
    if isinstance(obj, np.float32):
        return float(obj)
    if isinstance(obj, NestedMutableList) or isinstance(obj, list):
        return [convert(x) for x in obj]
    if isinstance(obj, NestedMutableDict) or isinstance(obj, dict):
        return {k: convert(v) for k, v in obj.items()}
    return obj


class JSONEncodedDict(TypeDecorator):
    impl = JSON

    def process_bind_param(self, value, dialect):
        return json.dumps(convert(value))

    def process_result_value(self, value, dialect):
        if value is not None:
            return json.loads(value)


class TrackTrackRelationshipDB(Base):
    __tablename__ = "track_track_relationships"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    source_id = Column(UUID(as_uuid=True), ForeignKey("tracks.id"))
    target_id = Column(UUID(as_uuid=True), ForeignKey("tracks.id"))
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(
        DateTime(timezone=True), server_default=func.now(), onupdate=func.now()
    )
    relationship_type = Column(String)
    meta = Column(mutable_json_type(dbtype=JSONEncodedDict, nested=True))

    # relationships
    source = relationship(
        "NendoTrackDB",
        foreign_keys=[source_id],
        back_populates="related_tracks",
    )
    target = relationship("NendoTrackDB", foreign_keys=[target_id])


class TrackCollectionRelationshipDB(Base):
    __tablename__ = "track_collection_relationships"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    source_id = Column(UUID(as_uuid=True), ForeignKey("tracks.id"))
    target_id = Column(UUID(as_uuid=True), ForeignKey("collections.id"))
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
    )
    relationship_type = Column(String)
    meta = Column(mutable_json_type(dbtype=JSONEncodedDict, nested=True))
    relationship_position = Column(Integer, nullable=False)

    # relationships
    source = relationship("NendoTrackDB", foreign_keys=[source_id])
    target = relationship("NendoCollectionDB", foreign_keys=[target_id])


class CollectionCollectionRelationshipDB(Base):
    __tablename__ = "collection_collection_relationships"
    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    source_id = Column(UUID(as_uuid=True), ForeignKey("collections.id"))
    target_id = Column(UUID(as_uuid=True), ForeignKey("collections.id"))
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
    )
    relationship_type = Column(String)
    meta = Column(mutable_json_type(dbtype=JSONEncodedDict, nested=True))

    # relationships
    source = relationship("NendoCollectionDB", foreign_keys=[source_id])
    target = relationship("NendoCollectionDB", foreign_keys=[target_id])


class NendoPluginDataDB(Base):
    __tablename__ = "plugin_data"

    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    track_id = Column(UUID(as_uuid=True), ForeignKey("tracks.id"))
    user_id = Column(UUID(as_uuid=True))  # , ForeignKey("users.id"))
    plugin_name = Column(String, nullable=False)
    plugin_version = Column(String, nullable=False)
    key = Column(String, nullable=False)
    value = Column(String, nullable=True)
    created_at = Column(DateTime(timezone=True), server_default=func.now())
    updated_at = Column(
        DateTime(timezone=True),
        server_default=func.now(),
        onupdate=func.now(),
    )

    # relationships
    # track = relationship("NendoTrackDB", backref="plugin_data")


class NendoTrackDB(Base):
    __tablename__ = "tracks"

    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    user_id = Column(UUID(as_uuid=True))
    track_type = Column(String, default="track")
    visibility = Column(ENUM(schema.Visibility), default="private")
    updated_at = Column(
        DateTime(timezone=True), default=func.now(), onupdate=func.now()
    )
    created_at = Column(DateTime(timezone=True), default=func.now())
    images = Column(mutable_json_type(dbtype=JSONEncodedDict, nested=True))
    resource = Column(mutable_json_type(dbtype=JSONEncodedDict, nested=True))
    meta = Column(mutable_json_type(dbtype=JSONEncodedDict, nested=True))

    # Relationships
    related_tracks = relationship(
        "TrackTrackRelationshipDB",
        primaryjoin="NendoTrackDB.id==TrackTrackRelationshipDB.source_id",
        foreign_keys="[TrackTrackRelationshipDB.source_id]",
        back_populates="source",
        # cascade="all, delete-orphan",
    )
    related_collections = relationship(
        "TrackCollectionRelationshipDB",
        primaryjoin="NendoTrackDB.id==TrackCollectionRelationshipDB.source_id",
        foreign_keys="[TrackCollectionRelationshipDB.source_id]",
        back_populates="source",
        # cascade="all, delete-orphan",
    )
    plugin_data = relationship(
        "NendoPluginDataDB",
        primaryjoin="NendoTrackDB.id==NendoPluginDataDB.track_id",
        foreign_keys="[NendoPluginDataDB.track_id]",
        # cascade="all, delete-orphan",
    )


class NendoBlobDB(Base):
    __tablename__ = "blobs"

    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    user_id = Column(UUID(as_uuid=True))
    visibility = Column(ENUM(schema.Visibility), default="private")
    updated_at = Column(
        DateTime(timezone=True),
        default=func.now(),
        onupdate=func.now(),
    )
    created_at = Column(DateTime(timezone=True), default=func.now())
    resource = Column(mutable_json_type(dbtype=JSONEncodedDict, nested=True))


class NendoCollectionDB(Base):
    __tablename__ = "collections"

    id = Column(UUID(as_uuid=True), primary_key=True, default=uuid.uuid4)
    name = Column(String)
    user_id = Column(UUID(as_uuid=True))
    description = Column(Text, default="")
    collection_type = Column(String, default="generic")
    visibility = Column(ENUM(schema.Visibility), default="private")
    updated_at = Column(
        DateTime(timezone=True),
        default=func.now(),
        onupdate=func.now(),
    )
    created_at = Column(DateTime(timezone=True), default=func.now())
    meta = Column(mutable_json_type(dbtype=JSONEncodedDict, nested=True))

    # relationships
    related_tracks = relationship(
        "TrackCollectionRelationshipDB",
        primaryjoin="NendoCollectionDB.id==TrackCollectionRelationshipDB.target_id",
        foreign_keys="[TrackCollectionRelationshipDB.target_id]",
        back_populates="target",
        # cascade="all, delete-orphan",
    )

    related_collections = relationship(
        "CollectionCollectionRelationshipDB",
        primaryjoin="NendoCollectionDB.id==CollectionCollectionRelationshipDB.source_id",
        foreign_keys="[CollectionCollectionRelationshipDB.source_id]",
        back_populates="source",
        # cascade="all, delete-orphan",
    )



"""Schemas for common typed of nendo objects."""
from __future__ import annotations


class NendoError(Exception):
    """Base class for nendo errors."""

    def __init__(self, message):  # noqa: D107
        super().__init__(message)
        self.message = message


class NendoIdError(NendoError):
    """Nendo error with an additional details field."""

    def __init__(self, message, target_id):  # noqa: D107
        super().__init__(message)
        self.target_id = target_id


class NendoPluginError(NendoError):
    """Generic Plugin Error.

    Any custom plugin error should inherit from this.
    """


class NendoPluginConfigError(NendoPluginError):
    """Error related to the (mis)configuration of a plugin."""


class NendoPluginLoadingError(NendoPluginError):
    """Error related to loading of a plugin."""


class NendoPluginRuntimeError(NendoPluginError):
    """Generic plugin runtime error."""


class NendoLibraryError(NendoError):
    """Generic library error.

    All other library error inherit from this.
    """


class NendoStorageError(NendoLibraryError):
    """Generic storage error."""


class NendoResourceError(NendoError):
    """Error related to a Nendo resource."""

    def __init__(self, message, src):  # noqa: D107
        super().__init__(message)
        self.src = src

    def __str__(self):
        return f"Error with resource {self.src}. Details: {self.message}"


class NendoLibraryIdError(NendoLibraryError):
    """A library error that refers to an item with a specified id."""

    def __init__(self, message, target_id):  # noqa: D107
        super().__init__(message)
        self.target_id = target_id


class NendoUserNotFoundError(NendoLibraryIdError):
    """Error raised if a user was accessed that does not exist."""

    def __str__(self):
        return f"User with ID {self.target_id} not found."


class NendoTrackNotFoundError(NendoLibraryIdError):
    """Error raised if a track was accessed that does not exist."""

    def __str__(self):
        return f"Track with ID {self.target_id} not found."


class NendoBlobNotFoundError(NendoLibraryIdError):
    """Error raised if a blob was accessed that does not exist."""

    def __str__(self):
        return f"Blob with ID {self.target_id} not found."


class NendoCollectionNotFoundError(NendoLibraryIdError):
    """Error raised if a collection was accessed that does not exist."""

    def __str__(self):
        return f"Collection with ID {self.target_id} not found."


class NendoRelationshipNotFoundError(NendoLibraryError):
    """Error raised if a relationship does not exist beween two IDs."""

    def __init__(self, message, source_id, target_id):  # noqa: D107
        super().__init__(message)
        self.source_id = source_id
        self.target_id = target_id

    def __str__(self):
        return f"Relationship between {self.source_id} and {self.target_id} not found."


class NendoPluginDataNotFoundError(NendoLibraryIdError):
    """Error raised if a plugin data was accessed that does not exist."""

    def __str__(self):
        return f"Plugin data with ID {self.target_id} not found."


class NendoFileFormatError(NendoLibraryError):
    """Error raised if there is an error with the format of a processed file."""


class NendoMalformedIDError(NendoLibraryIdError):
    """Error raised if the given ID is malformed / does not represent a valid UUID."""

    def __str__(self):
        return f"Malformed ID {self.target_id}."


class NendoBucketNotFoundError(NendoStorageError):
    """A library error that refers to an item with a specified id."""

    def __init__(self, message, bucket_name):  # noqa: D107
        super().__init__(message)
        self.bucket_name = bucket_name

    def __str__(self):
        return f"Bucket with name {self.bucket_name} not found."



"""Settings used to configure nendo."""
from enum import Enum
from functools import lru_cache
from typing import List

from pydantic import Field
from pydantic_settings import BaseSettings


class Environment(Enum):
    """Enum used to denote the environment in which nendo is running."""

    LOCAL = "local"
    TEST = "test"
    REMOTE = "remote"


class NendoConfig(BaseSettings):
    """The basic nendo configuration object.

    Uses pydantic's settings management features. Refer to the
    [documentation page on configuration management](../usage/config.md) for
    a full list of configuration options.
    """

    environment: Environment = Field(default=Environment.LOCAL)
    log_level: str = Field(default="WARNING")
    log_file_path: str = Field(default="")
    plugins: List[str] = Field(default_factory=list)
    library_plugin: str = Field(default="default")
    library_path: str = Field(default="nendo_library")
    user_name: str = Field(default="nendo")
    user_id: str = Field(default="ffffffff-1111-2222-3333-1234567890ab")
    auto_resample: bool = Field(default=False)
    default_sr: int = Field(default=44100)
    copy_to_library: bool = Field(default=True)
    auto_convert: bool = Field(default=True)
    skip_duplicate: bool = Field(default=True)
    max_threads: int = Field(default=2)
    batch_size: int = Field(default=10)
    stream_mode: bool = Field(default=False)
    stream_chunk_size: int = Field(default=1)


@lru_cache()
def get_settings() -> NendoConfig:
    """Return the Nendo configuration, cached."""
    return NendoConfig()



# -*- encoding: utf-8 -*-
# ruff: noqa: F401
"""Definition of all Nendo Core classes and modules."""

from .core import (
    NendoBlob,
    NendoBlobBase,
    NendoBlobCreate,
    NendoCollection,
    NendoCollectionBase,
    NendoCollectionCreate,
    NendoCollectionSlim,
    NendoPlugin,
    NendoPluginData,
    NendoPluginDataBase,
    NendoPluginDataCreate,
    NendoPluginRegistry,
    NendoRelationship,
    NendoRelationshipBase,
    NendoRelationshipCreate,
    NendoResource,
    NendoStorage,
    NendoStorageLocalFS,
    NendoTrack,
    NendoTrackBase,
    NendoTrackCreate,
    NendoTrackSlim,
    NendoUser,
    NendoUserBase,
    NendoUserCreate,
    RegisteredNendoPlugin,
    ResourceLocation,
    Visibility,
)
from .exception import (
    NendoBlobNotFoundError,
    NendoBucketNotFoundError,
    NendoCollectionNotFoundError,
    NendoError,
    NendoFileFormatError,
    NendoIdError,
    NendoLibraryError,
    NendoLibraryIdError,
    NendoPluginConfigError,
    NendoPluginDataNotFoundError,
    NendoPluginError,
    NendoPluginLoadingError,
    NendoPluginRuntimeError,
    NendoRelationshipNotFoundError,
    NendoResourceError,
    NendoStorageError,
    NendoTrackNotFoundError,
    NendoUserNotFoundError,
)
from .plugin import (
    NendoAnalysisPlugin,
    NendoEffectPlugin,
    NendoGeneratePlugin,
    NendoLibraryPlugin,
)



# -*- encoding: utf-8 -*-
# ruff: noqa: S603, S607, PLW1510
"""Module implementing the NendoLibraryPlugin using SQLAlchemy.

This module implements an SQLAlchemy version of the NendoLibraryPlugin.
The plugin is not suitable to be used by itself but should be inherited from
when implementing a new backend for the Nendo Library using SQLAlchemy. An example
is given by the DuckDB implementation used as the default Nendo Library.
"""

from __future__ import annotations

import logging
import os
import pickle
import subprocess
import uuid
from contextlib import contextmanager
from datetime import datetime
from tempfile import NamedTemporaryFile
from typing import TYPE_CHECKING, Any, Dict, Iterator, List, Optional, Union

import librosa
import numpy as np
import soundfile as sf
from sqlalchemy import Float, and_, asc, desc, func, or_, true
from sqlalchemy.orm import Query, Session, joinedload, sessionmaker
from sqlalchemy.sql.expression import cast
from sqlalchemy.sql.sqltypes import Text
from tinytag import TinyTag

from nendo import schema
from nendo.library import model
from nendo.utils import AudioFileUtils, ensure_uuid, md5sum

if TYPE_CHECKING:
    from pydantic import DirectoryPath, FilePath

logger = logging.getLogger("nendo")


class SqlAlchemyNendoLibrary(schema.NendoLibraryPlugin):
    """Implementation of the `NendoLibraryPlugin` using SQLAlchemy."""

    iter_index: int = 0
    user: schema.NendoUser = None

    def __init__(  # noqa: D107
        self,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.iter_index = 0

    @property
    def default_user(self):
        """Default Nendo user."""
        return schema.NendoUser(
            id=uuid.UUID(self.config.user_id),
            name=self.config.user_name,
            password="nendo",  # noqa: S106
            email="info@okio.ai",
            verified=True,
        )

    @contextmanager
    def session_scope(self):
        """Provide a transactional scope around a series of operations."""
        session = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        try:
            yield session
            session.commit()
        except:
            session.rollback()
            raise
        finally:
            session.close()

    def _disconnect(self):
        """Close DuckDB."""
        self.db.close()

    def _ensure_user_uuid(self, user_id: Optional[Union[str, uuid.UUID]] = None):
        return ensure_uuid(user_id) if user_id is not None else self.user.id

    def _convert_plugin_data(self, value: Any, user_id: Optional[uuid.UUID] = None):
        # numpy matrices are stored as .npy blob and the blob ID is returned as
        # the plugin data value
        if isinstance(value, (np.ndarray, np.matrix)):
            with NamedTemporaryFile(suffix=".npy", delete=True) as tmpfile:
                np.save(tmpfile.name, value)
                blob = self.store_blob(file_path=tmpfile.name, user_id=user_id)
                return str(blob.id)
        try:
            return str(value)
        except Exception as e:  # noqa: BLE001
            logger.error("Failed to load plugin data: %s", e)

    # ==========================
    #
    # TRACK MANAGEMENT FUNCTIONS
    #
    # ==========================

    def __len__(self):
        """Obtain the number of tracks."""
        with self.session_scope() as session:
            return session.query(model.NendoTrackDB).count()

    def __iter__(self):
        return self

    def __next__(self):
        with self.session_scope() as session:
            query = session.query(model.NendoTrackDB)
            track = query.offset(self.iter_index).first()
            if track:
                self.iter_index += 1
                return schema.NendoTrack.model_validate(track)
            raise StopIteration

    def _create_track_from_file(
        self,
        file_path: FilePath,
        track_type: str = "track",
        copy_to_library: Optional[bool] = None,
        skip_duplicate: Optional[bool] = None,
        user_id: Optional[uuid.UUID] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> schema.NendoTrackCreate:
        """Create a NendoTrack from the file given by file_path.

        Args:
            file_path (Union[FilePath, str]): Path to the file to be added.
            track_type (str, optional): Type of the track. Defaults to "track".
            copy_to_library (bool, optional): Flag that specifies whether
                the file should be copied into the library directory.
                Defaults to None.
            skip_duplicate (bool, optional): Flag that specifies whether a
                file should be added that already exists in the library, based on its
                file checksum. Defaults to None.
            user_id (UUID, optional): ID of user adding the track.
            meta (dict, optional): Metadata to attach to the track upon adding.

        Returns:
            schema.NendoTrackCreate: The created NendoTrack.
        """
        if not os.path.isfile(file_path):
            raise schema.NendoResourceError("File not found", file_path)

        if not AudioFileUtils().is_supported_filetype(file_path):
            raise schema.NendoResourceError("Unsupported filetype", file_path)

        file_checksum = md5sum(file_path)

        # skip adding a duplicate based on config flag and hashsum of the file
        skip_duplicate = skip_duplicate or self.config.skip_duplicate
        if skip_duplicate:
            tracks = list(self.find_tracks(value=file_checksum))
            if len(tracks) > 0:
                return schema.NendoTrack.model_validate(tracks[0])

        meta = meta or {}

        # gather file metadata
        try:
            # extract ID3 tags
            tags = TinyTag.get(file_path)
            meta.update(tags.as_dict())
        except KeyError as e:
            logger.error("Failed extracting tags from file: %s, Error: %s", file_path, e)

        # convert and save to library
        copy_to_library = copy_to_library or self.config.copy_to_library
        if copy_to_library or (self.config.auto_convert and file_path.endswith(".mp3")):
            try:
                file_stats = os.stat(file_path)
                sr = None
                if self.config.auto_convert:
                    if file_path.endswith(".mp3"):
                        signal, sr = librosa.load(path=file_path, sr=None, mono=False)
                    else:
                        signal, sr = sf.read(file=file_path)
                    # resample to default rate if required
                    if self.config.auto_resample and sr != self.config.default_sr:
                        logger.info(
                            "Auto-converting to SR of %d",
                            self.config.default_sr,
                        )
                        signal = librosa.resample(
                            signal,
                            orig_sr=sr,
                            target_sr=self.config.default_sr,
                        )
                        sr = self.config.default_sr
                    # sf.write expects the channels in the second dimension of the
                    # signal array! Librosa.load() loads them into the first dimension
                    signal = np.transpose(signal) if signal.shape[0] <= 2 else signal
                    path_in_library = self.storage_driver.save_signal(
                        file_name=self.storage_driver.generate_filename(
                            filetype="wav",
                            user_id=user_id,
                        ),
                        signal=signal,
                        sr=sr,
                        user_id=str(user_id) if user_id else str(self.user.id),
                    )
                else:
                    # save file to storage in its original format
                    path_in_library = self.storage_driver.save_file(
                        file_name=self.storage_driver.generate_filename(
                            filetype=os.path.splitext(file_path)[1][1:],  # without dot
                            user_id=user_id,
                        ),
                        file_path=file_path,
                        user_id=str(user_id) if user_id else str(self.user.id),
                    )

                # save the parsed sample rate
                if sr is not None:
                    meta["sr"] = sr

                meta.update(
                    {
                        "original_filename": os.path.basename(file_path),
                        "original_filepath": os.path.dirname(file_path),
                        "original_size": file_stats.st_size,
                        "original_checksum": file_checksum,
                    },
                )
                location = self.storage_driver.get_driver_location()
            except Exception as e:  # noqa: BLE001
                raise schema.NendoLibraryError(
                    f"Error copying file to the library: {e}.",
                ) from e
        else:
            path_in_library = file_path
            location = schema.ResourceLocation.original

        resource = schema.NendoResource(
            file_path=self.storage_driver.get_file_path(
                src=path_in_library,
                user_id=str(user_id) if user_id else str(self.user.id),
            ),
            file_name=self.storage_driver.get_file_name(
                src=path_in_library,
                user_id=str(user_id) if user_id else str(self.user.id),
            ),
            resource_type="audio",
            location=location,
            meta=meta,
        )
        return schema.NendoTrackCreate(
            nendo_instance=self.nendo_instance,
            resource=resource.model_dump(),
            user_id=user_id or self.user.id,
            track_type=track_type,
            meta=meta or {},
        )

    def _create_track_from_signal(
        self,
        signal: np.ndarray,
        sr: int,
        track_type: str = "track",
        meta: Optional[Dict[str, Any]] = None,
        user_id: Optional[uuid.UUID] = None,
    ) -> schema.NendoTrackCreate:
        """Create a NendoTrack from the given signal.

        Args:
            signal (np.ndarray): The numpy array containing the audio signal.
            sr (int): Sample rate
            track_type (str): Track type.
            user_id (UUID, optional): The ID of the user adding the track.
            meta (Dict[str, Any], optional): Track metadata. Defaults to {}.

        Returns:
            schema.NendoTrackCreate: The created NendoTrack.
        """
        target_file = None
        # sf.write expects the channels in the second dimension of the
        # signal array! Librosa.load() loads them into the first dimension
        signal = np.transpose(signal) if signal.shape[0] <= 2 else signal
        try:
            if self.config.auto_resample and sr != self.config.default_sr:
                logger.info("Auto-converting to SR of %d", self.config.default_sr)
                signal = librosa.resample(
                    signal, orig_sr=sr, target_sr=self.config.default_sr
                )
                sr = self.config.default_sr
            target_file = self.storage_driver.save_signal(
                file_name=self.storage_driver.generate_filename(
                    filetype="wav",
                    user_id=str(user_id) if user_id else str(self.user.id),
                ),
                signal=signal,
                sr=sr,
                user_id=user_id,
            )
        except Exception as e:  # noqa: BLE001
            raise schema.NendoLibraryError(
                f"Failed writing file {target_file} to the library. Error: {e}."
            ) from None
        resource = schema.NendoResource(
            file_path=self.storage_driver.get_file_path(
                src=target_file,
                user_id=str(user_id) if user_id else str(self.user.id),
            ),
            file_name=self.storage_driver.get_file_name(
                src=target_file,
                user_id=str(user_id) if user_id else str(self.user.id),
            ),
            resource_type="audio",
            location=self.storage_driver.get_driver_location(),
        )
        # save sample rate
        meta = meta or {}
        meta["sr"] = sr
        return schema.NendoTrackCreate(
            resource=resource.model_dump(),
            user_id=user_id or self.user.id,
            track_type=track_type,
            meta=meta,
        )

    @schema.NendoPlugin.batch_process
    def _add_tracks_db(
        self,
        file_paths: List[FilePath],
        track_type: str = "track",
        copy_to_library: Optional[bool] = None,
        skip_duplicate: bool = True,
        user_id: Optional[uuid.UUID] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> List[schema.NendoTrack]:
        """Add multiple tracks to the library from their file_paths.

        Args:
            file_paths (List[FilePath]): A list of full file paths to add as tracks.
            track_type (str, optional): The track_type of the tracks to add.
                Defaults to "track".
            copy_to_library (bool, optional): Flag that specifies whether
                the file should be copied into the library directory.
                Defaults to None.
            skip_duplicate (bool, optional): Flag that specifies whether a
                file should be added that already exists in the library, based on its
                file checksum. Defaults to None.
            user_id (UUID, optional): ID of user adding the track.
            meta (dict, optional): Metadata to attach to the track upon adding.

        Returns:
            List[schema.NendoTrack]: A list containing all added NendoTracks.
        """
        create_list = []
        for fp in file_paths:
            try:
                create_track = self._create_track_from_file(
                    file_path=fp,
                    track_type=track_type,
                    copy_to_library=copy_to_library,
                    skip_duplicate=skip_duplicate,
                    user_id=user_id or self.user.id,
                    meta=meta,
                )
                create_list.append(create_track)
            except schema.NendoLibraryError as e:
                logger.error("Failed adding file %s. Error: %s", fp, e)
        with self.session_scope() as session:
            db_tracks = self._upsert_tracks_db(tracks=create_list, session=session)
            return [schema.NendoTrack.model_validate(t) for t in db_tracks]

    def _get_related_tracks_query(
        self,
        track_id: uuid.UUID,
        session: Session,
        user_id: Optional[uuid.UUID] = None,
    ) -> Query:
        """Get tracks with a relationship to the track with track_id from the DB.

        Args:
            track_id (UUID): ID of the track to be searched for.
            session (Session): Session to be used for the transaction.
            user_id (UUID, optional): The user ID to filter for.

        Returns:
            Query: The SQLAlchemy query object.
        """
        user_id = user_id or self.user.id
        return (
            session.query(model.NendoTrackDB)
            .join(
                model.TrackTrackRelationshipDB,
                model.NendoTrackDB.id == model.TrackTrackRelationshipDB.source_id,
            )
            .filter(
                and_(
                    model.NendoTrackDB.user_id == user_id,
                    or_(
                        model.TrackTrackRelationshipDB.target_id == track_id,
                        model.TrackTrackRelationshipDB.source_id == track_id,
                    ),
                ),
            )
        )

    def _upsert_track_track_relationship(
        self, relationship: schema.NendoRelationshipBase, session: Session,
    ) -> schema.NendoTrack:
        """Insert or replace a track-to-track relationship in the database.

        Args:
            relationship (schema.NendoRelationshipBase): The relationship to upsert.
            session (sqlalchemy.Session): Session object to commit to.

        Returns:
            schema.NendoTrack: The upserted NendoTrack.
        """
        if type(relationship) == schema.NendoRelationshipCreate:
            # creating new relationship
            db_rel = model.TrackTrackRelationshipDB(**relationship.model_dump())
            session.add(db_rel)
        else:
            # updating existing relationship
            db_rel = (
                session.query(model.TrackTrackRelationshipDB)
                .filter_by(id=relationship.id)
                .first()
            )
            db_rel.source_id = relationship.source_id
            db_rel.target_id = relationship.target_id
            db_rel.relationship_type = relationship.relationship_type
            db_rel.meta = relationship.meta
        session.commit()
        return db_rel

    def _upsert_track_db(
        self, track: schema.NendoTrackBase, session: Session,
    ) -> model.NendoTrackDB:
        """Create track in DB or update if it exists.

        Args:
            track (schema.NendoTrackBase): Track object to be created
            session (Session): Session to be used for the transaction

        Returns:
            model.NendoTrackDB: The ORM model object of the upserted track
        """
        if type(track) == schema.NendoTrackCreate:
            # create new track
            track_dict = track.model_dump()
            track_dict.pop("nendo_instance")
            db_track = model.NendoTrackDB(**track_dict)
            session.add(db_track)
        else:
            # update existing track
            db_track = (
                session.query(model.NendoTrackDB).filter_by(id=track.id).one_or_none()
            )
            if db_track is None:
                raise schema.NendoTrackNotFoundError("Track not found", id=track.id)
            db_track.user_id = track.user_id
            db_track.visibility = track.visibility
            db_track.resource = track.resource.model_dump()
            db_track.track_type = track.track_type
            db_track.images = track.images
            db_track.meta = track.meta
        session.commit()
        return db_track

    def _upsert_tracks_db(
        self, tracks: List[schema.NendoTrackCreate], session: Session,
    ) -> List[model.NendoTrackDB]:
        """Create multiple tracks in DB or update if it exists.

        Args:
            tracks (List[schema.NendoTrackCreate]): Track object to be created
            session (Session): Session to be used for the transaction

        Returns:
            List[model.NendoTrackDB]: The ORM model objects of the upserted tracks
        """
        db_tracks = []
        for track in tracks:
            track_dict = track.model_dump()
            track_dict.pop("nendo_instance")
            db_tracks.append(model.NendoTrackDB(**track_dict))
        session.add_all(db_tracks)
        session.commit()
        return db_tracks

    def _get_all_plugin_data_db(
        self,
        track_id: uuid.UUID,
        session: Session,
        user_id: Optional[uuid.UUID] = None,
    ) -> List[model.NendoPluginDataDB]:
        """Get all plugin data related to a track from the DB.

        Args:
            track_id (UUID): Track ID to get the related plugin data for.
            session (Session): SQLAlchemy session.
            user_id (UUID, optional): The user ID to filter for.

        Returns:
            List[model.NendoPluginDataDB]: List of nendo plugin data entries.
        """
        user_id = user_id or self.user.id
        with self.session_scope() as session:
            plugin_data_db = (
                session.query(model.NendoPluginDataDB)
                .filter(
                    and_(
                        model.NendoPluginDataDB.track_id == track_id,
                        model.NendoPluginDataDB.user_id == user_id,
                    ),
                )
                .all()
            )
            return [
                schema.NendoPluginData.model_validate(pdb) for pdb in plugin_data_db
            ]

    def _get_all_plugin_data_db_by_name(
        self,
        track_id: uuid.UUID,
        plugin_name: str,
        session: Session,
        user_id: Optional[uuid.UUID] = None,
    ) -> List[model.NendoPluginDataDB]:
        """Get all plugin data related to a track and a given pluginfrom the DB.

        Args:
            track_id (uuid.UUID): Track ID to get the related plugin data for.
            plugin_name (str): Name of the plugin to get related data for.
            session (Session): SQLAlchemy session.
            user_id (UUID, optional): The user ID to filter for.

        Returns:
            List[model.NendoPluginDataDB]: List of nendo plugin data entries.
        """
        user_id = user_id or self.user.id
        with self.session_scope() as session:
            plugin_data_db = (
                session.query(model.NendoPluginDataDB)
                .filter(
                    and_(
                        model.NendoPluginDataDB.track_id == track_id,
                        model.NendoPluginDataDB.plugin_name == plugin_name,
                        model.NendoPluginDataDB.user_id == user_id,
                    ),
                )
                .all()
            )
            return (
                schema.NendoPluginData.model_validate(plugin_data_db)
                if plugin_data_db is not None
                else None
            )

    def _get_latest_plugin_data_db(
        self,
        track_id: uuid.UUID,
        plugin_name: str,
        plugin_version: str,
        key: str,
        session: Session,
        user_id: Optional[uuid.UUID] = None,
    ) -> model.NendoPluginDataDB:
        """Get the latest plugin data for a given track id, plugin name and data key.

        Args:
            track_id (uuid.UUID):  Track ID to get the related plugin data for.
            plugin_name (str): Name of the plugin to get related data for.
            plugin_version (str): Version of the plugin to get related data for.
            key (str): Key by which to filter the plugin data.
            session (Session): SQLAlchemy session.
            user_id (UUID, optional): The user ID to filter for.

        Returns:
            model.NendoPluginDataDB: A single nendo plugin data entry.
        """
        user_id = user_id or self.user.id
        with self.session_scope() as session:
            plugin_data_db = (
                session.query(model.NendoPluginDataDB)
                .filter(
                    and_(
                        model.NendoPluginDataDB.track_id == track_id,
                        model.NendoPluginDataDB.plugin_name == plugin_name,
                        model.NendoPluginDataDB.plugin_version == plugin_version,
                        model.NendoPluginDataDB.key == key,
                        model.NendoPluginDataDB.user_id == user_id,
                    ),
                )
                .order_by(model.NendoPluginDataDB.updated_at.desc())
                .first()
            )
            return (
                schema.NendoPluginData.model_validate(plugin_data_db)
                if plugin_data_db is not None
                else None
            )

    def _insert_plugin_data_db(
        self, plugin_data: schema.NendoPluginDataCreate, session: Session,
    ) -> model.NendoPluginDataDB:
        db_plugin_data = model.NendoPluginDataDB(**plugin_data.model_dump())
        session.add(db_plugin_data)
        session.commit()
        return db_plugin_data

    def _update_plugin_data_db(
        self,
        existing_plugin_data: model.NendoPluginDataDB,
        plugin_data: schema.NendoPluginData,
        session: Session,
    ) -> model.NendoPluginDataDB:
        if existing_plugin_data is None:
            logger.error("Plugin data not found!")
            return None
        existing_plugin_data.key = plugin_data.key
        existing_plugin_data.value = plugin_data.value
        existing_plugin_data.user_id = plugin_data.user_id
        session.commit()
        return existing_plugin_data

    def add_track(
        self,
        file_path: Union[FilePath, str],
        track_type: str = "track",
        copy_to_library: Optional[bool] = None,
        skip_duplicate: Optional[bool] = None,
        user_id: Optional[uuid.UUID] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> schema.NendoTrack:
        """Add the track given by path to the library.

        Args:
            file_path (Union[FilePath, str]): Path to the file to be added.
            track_type (str, optional): Type of the track. Defaults to "track".
            copy_to_library (bool, optional): Flag that specifies whether
                the file should be copied into the library directory.
                Defaults to None.
            skip_duplicate (bool, optional): Flag that specifies whether a
                file should be added that already exists in the library, based on its
                file checksum. Defaults to None.
            user_id (UUID, optional): ID of user adding the track.
            meta (dict, optional): Metadata to attach to the track upon adding.

        Returns:
            schema.NendoTrack: The track that was added to the library.
        """
        skip_duplicate = skip_duplicate or self.config.skip_duplicate
        track = self._create_track_from_file(
            file_path=file_path,
            track_type=track_type,
            copy_to_library=copy_to_library,
            skip_duplicate=skip_duplicate,
            user_id=user_id or self.user.id,
            meta=meta,
        )
        if track is not None:
            with self.session_scope() as session:
                db_track = self._upsert_track_db(track=track, session=session)
                track = schema.NendoTrack.model_validate(db_track)
        return track

    def add_track_from_signal(
        self,
        signal: np.ndarray,
        sr: int,
        track_type: str = "track",
        user_id: Optional[uuid.UUID] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> schema.NendoTrack:
        """Add a track to the library that is described by the given signal.

        Args:
            signal (np.ndarray): The numpy array containing the audio signal.
            sr (int): Sample rate
            track_type (str): Track type.
            user_id (UUID, optional): The ID of the user adding the track.
            meta (Dict[str, Any], optional): Track metadata. Defaults to {}.

        Returns:
            schema.NendoTrack: The added NendoTrack
        """
        track = self._create_track_from_signal(
            signal=signal,
            sr=sr,
            track_type=track_type,
            meta=meta,
            user_id=user_id or self.user.id,
        )
        if track is not None:
            with self.session_scope() as session:
                db_track = self._upsert_track_db(track=track, session=session)
                track = schema.NendoTrack.model_validate(db_track)
        return track

    def add_tracks(
        self,
        path: Union[str, DirectoryPath],
        track_type: str = "track",
        user_id: Optional[Union[str, uuid.UUID]] = None,
        copy_to_library: Optional[bool] = None,
        skip_duplicate: bool = True,
    ) -> schema.NendoCollection:
        """Scan the provided path and upsert the information into the library.

        Args:
            path (Union[str, DirectoryPath]): Path to the directory to be scanned
            track_type (str, optional): Track type for the new tracks
            user_id (UUID, optional): The ID of the user adding the track.
            copy_to_library (bool): Copy and convert the data into the nendo Library?
            skip_duplicate (bool): Skip adding duplicates?

        Returns:
            tracks (list[NendoTrack]): The tracks that were added to the Library
        """
        user_id = self._ensure_user_uuid(user_id)
        file_list = []
        if not os.path.exists(path):
            raise schema.NendoLibraryError(f"Source directory {path} does not exist.")
        for root, _, files in os.walk(path):
            file_list.extend(
                [
                    os.path.join(root, file)
                    for file in files
                    if AudioFileUtils().is_supported_filetype(file)
                ],
            )
        tracks = self._add_tracks_db(
            file_paths=file_list,
            track_type=track_type,
            copy_to_library=copy_to_library,
            skip_duplicate=skip_duplicate,
            user_id=user_id,
        )
        return self.add_collection(name=path, track_ids=[t.id for t in tracks])

    def add_track_relationship(
        self,
        track_one_id: Union[str, uuid.UUID],
        track_two_id: Union[str, uuid.UUID],
        relationship_type: str = "relationship",
        meta: Optional[Dict[str, Any]] = None,
    ) -> bool:
        """Add a relationship between two tracks."""
        track_one = self.get_track(track_one_id)  # track
        track_two = self.get_track(track_two_id)  # related

        # check that tracks are not the same
        if track_one_id == track_two_id:
            logger.error("Error must provide two different existing track ids")
            return False

        # check if tracks exist
        if track_one is None:
            logger.error("Error track id %s not found", str(track_one_id))
            return False

        if track_two is None:
            logger.error("Error track id %s not found", str(track_two_id))
            return False

        # create bidirectional relationship
        relationship_from = schema.NendoRelationshipCreate(
            source_id=track_one.id,
            target_id=track_two.id,
            relationship_type=relationship_type,
            meta=meta or {},
        )
        relationship_to = schema.NendoRelationshipCreate(
            source_id=track_two.id,
            target_id=track_one.id,
            relationship_type=relationship_type,
            meta=meta or {},
        )
        with self.session_scope() as session:
            relationship_from = schema.NendoRelationship.model_validate(
                self._upsert_track_track_relationship(
                    relationship=relationship_from, session=session,
                ),
            )
            _ = schema.NendoRelationship.model_validate(
                self._upsert_track_track_relationship(
                    relationship=relationship_to, session=session,
                ),
            )
        # avoid redundancy; only append one direction
        track_one.related_tracks.append(relationship_from)
        return True

    def add_related_track(
        self,
        file_path: Union[str, FilePath],
        related_track_id: Union[str, uuid.UUID],
        track_type: str = "track",
        user_id: Optional[Union[str, uuid.UUID]] = None,
        track_meta: Optional[Dict[str, Any]] = None,
        relationship_type: str = "relationship",
        meta: Optional[Dict[str, Any]] = None,
    ) -> schema.NendoTrack:
        """Add a track from a file with a relationship to another track."""
        user_id = self._ensure_user_uuid(user_id=user_id)
        related_track_id = ensure_uuid(related_track_id)
        track = self.add_track(
            file_path=file_path, track_type=track_type, user_id=user_id, meta=track_meta,
        )
        # create bidirectional relationship
        relationship_from = schema.NendoRelationshipCreate(
            source_id=track.id,
            target_id=related_track_id,
            relationship_type=relationship_type,
            meta=meta or {},
        )
        relationship_to = schema.NendoRelationshipCreate(
            source_id=related_track_id,
            target_id=track.id,
            relationship_type=relationship_type,
            meta=meta or {},
        )
        with self.session_scope() as session:
            relationship_from = schema.NendoRelationship.model_validate(
                self._upsert_track_track_relationship(
                    relationship=relationship_from, session=session,
                ),
            )
            _ = schema.NendoRelationship.model_validate(
                self._upsert_track_track_relationship(
                    relationship=relationship_to, session=session,
                ),
            )
        # avoid redundancy; only append one direction
        track.related_tracks.append(relationship_from)
        return track

    def add_related_track_from_signal(
        self,
        signal: np.ndarray,
        sr: int,
        related_track_id: Union[str, uuid.UUID],
        track_type: str = "track",
        user_id: Optional[uuid.UUID] = None,
        track_meta: Optional[Dict[str, Any]] = None,
        relationship_type: str = "relationship",
        meta: Optional[Dict[str, Any]] = None,
    ) -> schema.NendoTrack:
        """Add a track from a signal with a relationship to another track."""
        user_id = user_id or self.user.id
        related_track_id = ensure_uuid(related_track_id)
        track = self.add_track_from_signal(
            signal=signal, sr=sr, track_type=track_type, meta=track_meta,
        )
        # create bidirectional relationship
        relationship_from = schema.NendoRelationshipCreate(
            source_id=track.id,
            target_id=related_track_id,
            relationship_type=relationship_type,
            meta=meta or {},
        )
        relationship_to = schema.NendoRelationshipCreate(
            source_id=track.id,
            target_id=related_track_id,
            relationship_type=relationship_type,
            meta=meta or {},
        )
        with self.session_scope() as session:
            relationship_from = schema.NendoRelationship.model_validate(
                self._upsert_track_track_relationship(
                    relationship=relationship_from, session=session,
                ),
            )
            _ = schema.NendoRelationship.model_validate(
                self._upsert_track_track_relationship(
                    relationship=relationship_to, session=session,
                ),
            )
        # avoid redundancy; only append one direction
        track.related_tracks.append(relationship_from)
        return track

    def update_track(
        self,
        track: schema.NendoTrack,
    ) -> schema.NendoTrack:
        """Updates the given track by storing it to the database.

        Args:
            track (NendoTrack): The track to be stored to the database.

        Raises:
            NendoTrackNotFoundError: If the track passed to the function
                does not exist in the database.

        Returns:
            NendoTrack: The updated track.
        """
        with self.session_scope() as session:
            # update existing track
            db_track = (
                session.query(model.NendoTrackDB).filter_by(id=track.id).one_or_none()
            )
            if db_track is None:
                raise schema.NendoTrackNotFoundError("Track not found", id=track.id)
            db_track.user_id = track.user_id
            db_track.visibility = track.visibility
            db_track.resource = track.resource.model_dump()
            db_track.track_type = track.track_type
            db_track.images = track.images
            db_track.meta = track.meta
            session.commit()
        return db_track

    def add_plugin_data(
        self,
        track_id: Union[str, uuid.UUID],
        plugin_name: str,
        plugin_version: str,
        key: str,
        value: Any,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        replace: bool = False,
    ) -> schema.NendoPluginData:
        """Add plugin data to a NendoTrack and persist changes into the DB.

        Args:
            track_id (Union[str, UUID]): ID of the track to which
                the plugin data should be added.
            plugin_name (str): Name of the plugin.
            plugin_version (str): Version of the plugin.
            key (str): Key under which to save the data.
            value (Any): Data to save.
            user_id (Union[str, UUID], optional): ID of user adding the plugin data.
            replace (bool, optional): Flag that determines whether
                the last existing data point for the given plugin name and -version
                is overwritten or not. Defaults to False.

        Returns:
            NendoPluginData: The saved plugin data as a NendoPluginData object.
        """
        # create plugin data
        user_id = self._ensure_user_uuid(user_id)
        value_converted = self._convert_plugin_data(value=value, user_id=user_id)
        plugin_data = schema.NendoPluginDataCreate(
            track_id=ensure_uuid(track_id),
            user_id=ensure_uuid(user_id),
            plugin_name=plugin_name,
            plugin_version=plugin_version,
            key=key,
            value=value_converted,
        )
        with self.session_scope() as session:
            if replace:
                existing_plugin_data = self._get_latest_plugin_data_db(
                    track_id=track_id,
                    plugin_name=plugin_name,
                    plugin_version=plugin_version,
                    key=key,
                    session=session,
                )

                if existing_plugin_data is not None:
                    db_plugin_data = self._update_plugin_data_db(
                        existing_plugin_data=existing_plugin_data,
                        plugin_data=plugin_data,
                        session=session,
                    )
                else:
                    db_plugin_data = self._insert_plugin_data_db(
                        plugin_data=plugin_data, session=session,
                    )
            else:
                db_plugin_data = self._insert_plugin_data_db(
                    plugin_data=plugin_data, session=session,
                )
            return schema.NendoPluginData.model_validate(db_plugin_data)

    def get_track(self, track_id: uuid.UUID) -> schema.NendoTrack:
        """Get a single track from the library by ID."""
        with self.session_scope() as session:
            track_db = (
                session.query(model.NendoTrackDB)
                .filter(model.NendoTrackDB.id == track_id)
                .one_or_none()
            )
            return (
                schema.NendoTrack.model_validate(track_db) if track_db is not None else None
            )

    @schema.NendoPlugin.stream_output
    def get_tracks(
        self,
        query: Optional[Query] = None,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: str = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        session: Optional[Session] = None,
    ) -> Union[List, Iterator]:
        """Get tracks based on the given query parameters.

        Args:
            query (Optional[Query]): Query object to build from.
            user_id (Union[str, UUID], optional): ID of user to filter tracks by.
            order_by (Optional[str]): Key used for ordering the results.
            order (Optional[str]): Order in which to retrieve results ("asc" or "desc").
            limit (Optional[int]): Limit the number of returned results.
            offset (Optional[int]): Offset into the paginated results (requires limit).
            session (sqlalchemy.Session): Session object to commit to.

        Returns:
            Union[List, Iterator]: List or generator of tracks, depending on the
                configuration variable stream_mode
        """
        user_id = self._ensure_user_uuid(user_id)
        s = session or self.session_scope()
        with s as session_local:
            if query:
                query_local = query
            else:
                query_local = session_local.query(model.NendoTrackDB).filter(
                    model.NendoTrackDB.user_id == user_id,
                )
            if order_by:
                if order_by == "random":
                    query_local = query_local.order_by(func.random())
                elif order == "desc":
                    query_local = query_local.order_by(
                        desc(getattr(model.NendoTrackDB, order_by)),
                    )
                else:
                    query_local = query_local.order_by(
                        asc(getattr(model.NendoTrackDB, order_by)),
                    )
            if limit:
                query_local = query_local.limit(limit)
                if offset:
                    query_local = query_local.offset(offset)

            if self.config.stream_chunk_size > 1:
                chunk = []
                for track in query_local:
                    chunk.append(schema.NendoTrack.model_validate(track))
                    if len(chunk) == self.config.stream_chunk_size:
                        yield chunk
                        chunk = []
                if chunk:  # yield remaining tracks in non-full chunk
                    yield chunk
            else:
                for track in query_local:
                    yield schema.NendoTrack.model_validate(track)

    def get_related_tracks(
        self,
        track_id: Union[str, uuid.UUID],
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: Optional[str] = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[List, Iterator]:
        """Get tracks with a relationship to the track with track_id.

        Args:
            track_id (str): ID of the track to be searched for.
            user_id (Union[str, UUID], optional): The user ID to filter for.
            order_by (Optional[str]): Key used for ordering the results.
            order (Optional[str]): Order in which to retrieve results ("asc" or "desc").
            limit (Optional[int]): Limit the number of returned results.
            offset (Optional[int]): Offset into the paginated results (requires limit).

        Returns:
            Union[List, Iterator]: List or generator of tracks, depending on the
                configuration variable stream_mode
        """
        user_id = self._ensure_user_uuid(user_id)
        with self.session_scope() as session:
            query = self._get_related_tracks_query(
                track_id=ensure_uuid(track_id),
                session=session,
                user_id=user_id,
            )
            return self.get_tracks(
                query=query,
                order_by=order_by,
                order=order,
                limit=limit,
                offset=offset,
                session=session,
            )

    def find_tracks(
        self,
        value: str,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: str = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[List, Iterator]:
        """Obtain tracks from the db by fulltext search.

        Args:
            value (str): The value to search for. The value is matched against
                text representations of the track's `meta` and `resource` fields.
            user_id (Union[str, UUID], optional): The user ID to filter for.
            order_by (Optional[str]): Key used for ordering the results.
            limit (Optional[int]): Limit the number of returned results.
            offset (Optional[int]): Offset into the paginated results (requires limit).

        Returns:
            Union[List, Iterator]: List or generator of tracks, depending on the
                configuration variable stream_mode
        """
        user_id = self._ensure_user_uuid(user_id)
        with self.session_scope() as session:
            query = session.query(model.NendoTrackDB).filter(
                and_(
                    or_(
                        cast(model.NendoTrackDB.resource, Text()).ilike(
                            "%{}%".format(value),
                        ),
                        cast(model.NendoTrackDB.meta, Text()).ilike(
                            "%{}%".format(value),
                        ),
                    ),
                    model.NendoTrackDB.user_id == user_id,
                ),
            )
            return self.get_tracks(
                query=query,
                order_by=order_by,
                order=order,
                limit=limit,
                offset=offset,
                session=session,
            )

    def filter_tracks(
        self,
        filters: Optional[Dict[str, Any]] = None,
        resource_filters: Optional[Dict[str, Any]] = None,
        track_type: Optional[Union[str, List[str]]] = None,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        collection_id: Optional[Union[str, uuid.UUID]] = None,
        plugin_names: Optional[List[str]] = None,
        order_by: Optional[str] = None,
        order: str = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        session: Optional[Session] = None,
    ) -> Union[List, Iterator]:
        """Obtain tracks from the db by filtering over plugin data.

        Args:
            filters (Optional[dict]): Dictionary containing the filters to apply.
                Defaults to None.
            resource_filters (dict): Dictionary containing the keywords to search for
                over the track.resource.meta field. The dictionary's values
                should contain singular search tokens and the keys currently have no
                effect but might in the future. Defaults to {}.
            track_type (Union[str, List[str]], optional): Track type to filter for.
                Can be a singular type or a list of types. Defaults to None.
            user_id (Union[str, UUID], optional): The user ID to filter for.
            collection_id (Union[str, uuid.UUID], optional): Collection id to
                which the filtered tracks must have a relationship. Defaults to None.
            plugin_names (list, optional): List used for applying the filter only to
                data of certain plugins. If None, all plugin data related to the track
                is used for filtering.
            order_by (str, optional): Key used for ordering the results.
            order (str, optional): Ordering ("asc" vs "desc"). Defaults to "asc".
            limit (int, optional): Limit the number of returned results.
            offset (int, optional): Offset into the paginated results (requires limit).

        Returns:
            Union[List, Iterator]: List or generator of tracks, depending on the
                configuration variable stream_mode
        """
        user_id = self._ensure_user_uuid(user_id)
        s = session or self.session_scope()
        with s as session_local:
            """Obtain tracks from the db by filtering w.r.t. various fields."""
            query = session_local.query(model.NendoTrackDB).filter(
                model.NendoTrackDB.user_id == user_id,
            )
            plugin_name_condition = true()
            if (
                plugin_names is not None
                and isinstance(plugin_names, list)
                and len(plugin_names) > 0
            ):
                plugin_name_condition = model.NendoPluginDataDB.plugin_name.in_(
                    plugin_names,
                )

            # apply track type filter if applicable
            if track_type is not None:
                if isinstance(track_type, list):
                    query = query.filter(model.NendoTrackDB.value.in_(track_type))
                else:
                    query = query.filter(model.NendoTrackDB.track_type == track_type)

            # apply collection filter if applicable
            if collection_id is not None:
                collection_id = ensure_uuid(collection_id)

                query = query.join(
                    model.TrackCollectionRelationshipDB,
                    model.NendoTrackDB.id
                    == model.TrackCollectionRelationshipDB.source_id,
                ).filter(
                    model.TrackCollectionRelationshipDB.target_id == collection_id,
                )

            # apply resource filters if applicable
            if resource_filters:
                values = [v for v in resource_filters.values() if isinstance(v, list)]
                or_filter = or_(
                    *(
                        cast(model.NendoTrackDB.resource, Text()).ilike(
                            "%{}%".format(str(value)),
                        )
                        for rf in values
                        for value in rf
                    ),
                )
                query = query.filter(or_filter)

            # apply plugin data filters
            if filters is not None:
                for k, v in filters.items():
                    if v is None:
                        continue

                    # range
                    if type(v) == tuple:
                        query = query.filter(
                            model.NendoTrackDB.plugin_data.any(
                                and_(
                                    plugin_name_condition,
                                    model.NendoPluginDataDB.key == k,
                                    cast(model.NendoPluginDataDB.value, Float)
                                    >= cast(v[0], Float),
                                    cast(model.NendoPluginDataDB.value, Float)
                                    <= cast(v[1], Float),
                                ),
                            ),
                        )
                    # multiselect
                    elif isinstance(v, list):
                        sv = [str(vi) for vi in v]
                        query = query.filter(
                            model.NendoTrackDB.plugin_data.any(
                                and_(
                                    plugin_name_condition,
                                    model.NendoPluginDataDB.key == k,
                                    model.NendoPluginDataDB.value.in_(sv),
                                ),
                            ),
                        )
                    # fuzzy match
                    else:
                        query = query.filter(
                            model.NendoTrackDB.plugin_data.any(
                                and_(
                                    plugin_name_condition,
                                    model.NendoPluginDataDB.key == k,
                                    cast(model.NendoPluginDataDB.value, Text()).ilike(
                                        "%{}%".format(str(v)),
                                    ),
                                ),
                            ),
                        )

            return self.get_tracks(
                query=query,
                order_by=order_by,
                order=order,
                limit=limit,
                offset=offset,
                session=session,
            )

    def remove_track(
        self,
        track_id: Union[str, uuid.UUID],
        remove_relationships: bool = False,
        remove_plugin_data: bool = True,
        remove_resources: bool = True,
        user_id: Optional[Union[str, uuid.UUID]] = None,
    ) -> bool:
        """Delete track from library by ID.

        Args:
            track_id (Union[str, uuid.UUID]): The ID of the track to remove
            remove_relationships (bool):
                If False prevent deletion if related tracks exist,
                if True delete relationships together with the object
            remove_plugin_data (bool):
                If False prevent deletion if related plugin data exist
                if True delete plugin data together with the object
            remove_resources (bool):
                If False, keep the related resources, e.g. files
                if True, delete the related resources
            user_id (Union[str, UUID], optional): The ID of the user owning the track.

        Returns:
            success (bool): True if removal was successful, False otherwise.
        """
        user_id = self._ensure_user_uuid(user_id)
        track_id = ensure_uuid(track_id)
        with self.session_scope() as session:
            tracks_with_relations = self._get_related_tracks_query(
                track_id=track_id, session=session, user_id=user_id,
            ).all()
            collections_with_relations = (
                session.query(model.NendoCollectionDB)
                .join(
                    model.TrackCollectionRelationshipDB,
                    model.NendoCollectionDB.id
                    == model.TrackCollectionRelationshipDB.target_id,
                )
                .filter(model.TrackCollectionRelationshipDB.source_id == track_id)
                .all()
            )
            related_plugin_data = self._get_all_plugin_data_db(
                track_id=track_id, session=session,
            )
            if len(related_plugin_data) > 0:
                if remove_plugin_data:
                    logger.info("Removing %d plugin data", len(related_plugin_data))
                    session.query(model.NendoPluginDataDB).filter(
                        model.NendoPluginDataDB.track_id == track_id,
                    ).delete()
                    session.commit()
                else:
                    logger.warning(
                        "Cannot remove due to %d existing "
                        "plugin data entries. Set `remove_plugin_data=True` "
                        "to remove them.", len(related_plugin_data),
                    )
                    return False
            n_rel = len(tracks_with_relations) + len(collections_with_relations)
            if n_rel > 0:
                if remove_relationships:
                    session.query(model.TrackTrackRelationshipDB).filter(
                        model.TrackTrackRelationshipDB.target_id == track_id,
                    ).delete()
                    session.query(model.TrackTrackRelationshipDB).filter(
                        model.TrackTrackRelationshipDB.source_id == track_id,
                    ).delete()
                    # remove track from all collections
                    for collection in collections_with_relations:
                        self._remove_track_from_collection_db(
                            track_id=track_id,
                            collection_id=collection.id,
                            session=session,
                        )
                    session.commit()
                else:
                    logger.warning(
                        "Cannot remove due to %s existing relationships. "
                        "Set `remove_relationships=True` to remove them.",
                        n_rel,
                    )
                    return False
            # delete actual target
            target = (
                session.query(model.NendoTrackDB)
                .filter(model.NendoTrackDB.id == track_id)
                .first()
            )
            session.delete(target)
        # only delete if file has been copied to the library
        # ("original_filepath" is present)
        if (
            remove_resources
            and "original_filepath"
            in schema.NendoTrack.model_validate(target).resource.meta
        ):
            logger.info("Removing resources associated with Track %s", str(track_id))
            return self.storage_driver.remove_file(
                file_name=target.resource["file_name"],
                user_id=str(user_id) if user_id else str(self.user.id),
            )
        return True

    def export_track(
        self,
        track_id: Union[str, uuid.UUID],
        file_path: str,
        file_format: str = "wav",
    ) -> str:
        """Export the track to a file.

        Args:
            track_id (Union[str, uuid.UUID]): The ID of the target track to export.
            file_path (str): Path to the exported file. Can be either a full
                file path or a directory path. If a directory path is given,
                a filename will be automatically generated and the file will be
                exported to the format specified as file_format. If a full file
                path is given, the format will be deduced from the path and the
                file_format parameter will be ignored.
            file_format (str, optional): Format of the exported track. Ignored if
                file_path is a full file path. Defaults to "wav".

        Returns:
            str: The path to the exported file.
        """
        track = self.get_track(track_id=track_id)
        # Check if file_path is a directory
        if os.path.isdir(file_path):
            # Generate a filename with timestamp
            if track.has_meta("original_filename"):
                original_filename = track.get_meta("original_filename")
            else:
                original_filename = track.resource.file_name
            file_name = (
                f"{original_filename}_nendo_"
                f"{datetime.now().strftime('%Y%m%d%H%M%S')}" # noqa: DTZ005
                f".{file_format}"
            )
            file_path = os.path.join(file_path, file_name)
        else:
            # Deduce file format from file extension
            file_format = os.path.splitext(file_path)[1].lstrip(".")

        # Exporting the audio
        temp_path = None
        signal = track.signal
        signal = np.transpose(signal) if signal.shape[0] <= 2 else signal
        if file_format in ("wav", "ogg"):
            sf.write(file_path, signal, track.sr, format=file_format)
        elif file_format == "mp3":
            # Create a temporary WAV file for conversion
            temp_path = file_path.rsplit(".", 1)[0] + ".wav"
            sf.write(temp_path, signal, track.sr)
            subprocess.run(
                ["ffmpeg", "-i", temp_path, "-acodec", "libmp3lame", file_path],
            )
        else:
            raise ValueError(
                "Unsupported file format. Supported formats are 'wav', 'mp3', and 'ogg'.",
            )

        # Clean up temporary file if used
        if temp_path and os.path.exists(temp_path):
            os.remove(temp_path)

        return file_path

    # ===============================
    #
    # COLLECTION MANAGEMENT FUNCTIONS
    #
    # ===============================

    def _get_related_collections_query(
        self,
        collection_id: uuid.UUID,
        session: Session,
        user_id: Optional[uuid.UUID] = None,
    ) -> Query:
        """Create a query for the collections related to a given collection.

        Args:
            collection_id (UUID): ID of the collection to be searched for.
            session (Session): Session to be used for the transaction.
            user_id (UUID, optional): The user ID to filter for.

        Returns:
            Query: The SQLAlchemy query object.
        """
        user_id = user_id or self.user.id
        return (
            session.query(model.NendoCollectionDB)
            .filter(model.NendoCollectionDB.user_id == user_id)
            .options(
                joinedload(model.NendoCollectionDB.related_tracks).joinedload(
                    model.TrackCollectionRelationshipDB.source,
                ),
            )
            .join(
                model.CollectionCollectionRelationshipDB,
                model.NendoCollectionDB.id
                == model.CollectionCollectionRelationshipDB.source_id,
            )
            .filter(
                model.CollectionCollectionRelationshipDB.target_id == collection_id,
            )
        )

    def _remove_track_from_collection_db(
        self,
        track_id: uuid.UUID,
        collection_id: uuid.UUID,
        session: Session,
    ) -> bool:
        """Deletes a relationship from the track to the collection in the db.

        Args:
            collection_id (uuid.UUID): Collection id.
            track_id (uuid.UUID): Track id.
            session (sqlalchemy.Session): Session object

        Returns:
            success (bool): True if removal was successful, False otherwise.
        """
        target_relationship = (
            session.query(model.TrackCollectionRelationshipDB)
            .filter(
                and_(
                    model.TrackCollectionRelationshipDB.source_id == track_id,
                    model.TrackCollectionRelationshipDB.target_id == collection_id,
                ),
            )
            .first()
        )
        if target_relationship is None:
            raise schema.NendoRelationshipNotFoundError(
                "Relationship not found", track_id, collection_id,
            )
        session.delete(target_relationship)
        # Adjust positions of other states
        rel_db = model.TrackCollectionRelationshipDB
        session.query(model.TrackCollectionRelationshipDB).filter(
            rel_db.target_id == collection_id,
            rel_db.relationship_position
            > target_relationship.relationship_position,
        ).update({rel_db.relationship_position: rel_db.relationship_position - 1})

    def add_collection(
        self,
        name: str,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        track_ids: Optional[List[Union[str, uuid.UUID]]] = None,
        description: str = "",
        collection_type: str = "collection",
        visibility: schema.Visibility = schema.Visibility.private,
        meta: Optional[Dict[str, Any]] = None,
    ) -> schema.NendoCollection:
        """Creates a new collection and saves it into the DB.

        Args:
            track_ids (List[Union[str, uuid.UUID]]): List of track ids
                to be added to the collection.
            name (str): Name of the collection.
            user_id (UUID, optional): The ID of the user adding the collection.
            description (str): Description of the collection.
            collection_type (str): Type of the collection.
            meta (Dict[str, Any]): Metadata of the collection.

        Returns:
            schema.NendoCollection: The newly created NendoCollection object.
        """
        user_id = self._ensure_user_uuid(user_id)
        if track_ids is None:
            track_ids = []
        with self.session_scope() as session:
            # Fetch the track objects
            track_objs = (
                session.query(model.NendoTrackDB)
                .filter(
                    and_(
                        model.NendoTrackDB.id.in_(
                            [
                                uuid.UUID(t) if isinstance(t, str) else t
                                for t in track_ids
                            ],
                        ),
                        model.NendoTrackDB.user_id == user_id,
                    ),
                )
                .all()
            )

            # Create a new collection object
            new_collection = model.NendoCollectionDB(
                name=name,
                user_id=user_id,
                description=description,
                collection_type=collection_type,
                visibility=visibility,
                meta=meta or {},
            )
            session.add(new_collection)
            session.commit()
            session.refresh(new_collection)
            # Create relationships from tracks to collection
            for idx, track in enumerate(track_objs):
                tc_relationship = model.TrackCollectionRelationshipDB(
                    source_id=track.id,
                    target_id=new_collection.id,
                    relationship_type="track",
                    meta={},
                    relationship_position=idx,
                )
                session.add(tc_relationship)
            session.commit()
            session.refresh(new_collection)

            return schema.NendoCollection.model_validate(new_collection)

    def add_related_collection(
        self,
        track_ids: List[Union[str, uuid.UUID]],
        collection_id: Union[str, uuid.UUID],
        name: str,
        description: str = "",
        user_id: Optional[Union[str, uuid.UUID]] = None,
        relationship_type: str = "relationship",
        meta: Optional[Dict[str, Any]] = None,
    ) -> schema.NendoCollection:
        """Adds a new collection with a relationship to the given collection.

        Args:
            track_ids (List[Union[str, uuid.UUID]]): List of track ids.
            collection_id (Union[str, uuid.UUID]): Existing collection id.
            name (str): Name of the new related collection.
            description (str): Description of the new related collection.
            user_id (UUID, optional): The ID of the user adding the collection.
            relationship_type (str): Type of the relationship.
            meta (Dict[str, Any]): Meta of the new related collection.

        Returns:
            schema.NendoCollection: The newly added NendoCollection object.
        """
        user_id = self._ensure_user_uuid(user_id)
        # Create a new collection
        new_collection = self.add_collection(
            name=name,
            user_id=user_id,
            track_ids=track_ids,
            description=description,
            collection_type=relationship_type,
            meta=meta,
        )

        with self.session_scope() as session:
            if isinstance(collection_id, str):
                collection_id = uuid.UUID(collection_id)
            collection = (
                session.query(model.NendoCollectionDB)
                .filter_by(id=collection_id)
                .first()
            )

            # Check if the collection does not exist
            if not collection:
                raise schema.NendoCollectionNotFoundError(
                    "Collection not found", id=collection_id,
                )

            # create bidirectional relationship
            relationship_from = schema.NendoRelationshipCreate(
                source_id=new_collection.id,
                target_id=collection_id,
                relationship_type=relationship_type,
                meta=meta or {},
            )
            relationship_to = schema.NendoRelationshipCreate(
                source_id=collection_id,
                target_id=new_collection.id,
                relationship_type=relationship_type,
                meta=meta or {},
            )
            relationship_from = model.CollectionCollectionRelationshipDB(
                **relationship_from.model_dump(),
            )
            relationship_to = model.CollectionCollectionRelationshipDB(
                **relationship_to.model_dump(),
            )
            session.add(relationship_from)
            session.add(relationship_to)

        new_collection.related_collections.append(relationship_from)
        return new_collection

    def add_track_to_collection(
        self,
        track_id: Union[str, uuid.UUID],
        collection_id: Union[str, uuid.UUID],
        position: Optional[int] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> schema.NendoCollection:
        """Creates a relationship from the track to the collection.

        Args:
            track_id (Union[str, uuid.UUID]): ID of the track to add.
            collection_id (Union[str, uuid.UUID]): ID of the collection to
                which to add the track.
            position (int, optional): Target position of the track inside
                the collection.
            meta (Dict[str, Any]): Metadata of the relationship.

        Returns:
            schema.NendoCollection: The updated NendoCollection object.
        """
        with self.session_scope() as session:
            # Convert IDs to UUIDs if they're strings
            if isinstance(collection_id, str):
                collection_id = uuid.UUID(collection_id)
            if isinstance(track_id, str):
                track_id = uuid.UUID(track_id)

            # Check the collection and track objects
            collection = (
                session.query(model.NendoCollectionDB)
                .filter_by(id=collection_id)
                .first()
            )
            track = session.query(model.NendoTrackDB).filter_by(id=track_id).first()
            if not collection:
                raise schema.NendoCollectionNotFoundError(
                    "The collection does not exist", collection_id,
                )
            if not track:
                raise schema.NendoCollectionNotFoundError(
                    "The track does not exist", track_id,
                )

            rc_rel_db = model.TrackCollectionRelationshipDB
            if position is not None:
                # Update other states to keep ordering consistent
                session.query(rc_rel_db).filter(
                    rc_rel_db.target_id == collection_id,
                    rc_rel_db.relationship_position >= position,
                ).update(
                    {
                        rc_rel_db.relationship_position: rc_rel_db.relationship_position
                        + 1,
                    },
                )
            else:
                # If no position specified, add at the end
                last_state = (
                    session.query(model.TrackCollectionRelationshipDB)
                    .filter_by(target_id=collection_id)
                    .order_by(
                        model.TrackCollectionRelationshipDB.relationship_position.desc(),
                    )
                    .first()
                )
                position = last_state.relationship_position + 1 if last_state else 0

            # Create a relationship from the track to the collection
            tc_relationship = model.TrackCollectionRelationshipDB(
                source_id=track_id,
                target_id=collection_id,
                relationship_type="track",
                meta=meta or {},
                relationship_position=position,
            )
            session.add(tc_relationship)
            session.commit()
            session.refresh(collection)
            return schema.NendoCollection.model_validate(collection)

    def get_collection_tracks(
        self, collection_id: uuid.UUID,
    ) -> List[schema.NendoTrack]:
        """Get all tracks from a collection.

        Args:
            collection_id (uuid.UUID): ID of the collection to get the tracks from.

        Returns:
            List[schema.NendoTrack]: List of tracks in the collection.
        """
        with self.session_scope() as session:
            tracks_db = (
                session.query(model.NendoTrackDB)
                .join(
                    model.TrackCollectionRelationshipDB,
                    model.TrackCollectionRelationshipDB.source_id
                    == model.NendoTrackDB.id,
                )
                .options(joinedload(model.NendoTrackDB.related_collections))
                .filter(model.TrackCollectionRelationshipDB.target_id == collection_id)
                .all()
            )

            return (
                [schema.NendoTrack.model_validate(t) for t in tracks_db]
                if tracks_db is not None
                else None
            )

    def get_collection(
        self, collection_id: uuid.UUID, details: bool = True,
    ) -> Union[schema.NendoCollection, schema.NendoCollectionSlim]:
        """Get a collection by its ID.

        Args:
            collection_id (uuid.UUID): ID of the target collection.
            details (bool, optional): Flag that defines whether the result should
                contain all fields or only a subset. Defaults to True.

        Returns:
            Union[NendoCollection, NendoCollectionSlim]: Collection object, compact
                version if the `details` flag has been set to False.
        """
        with self.session_scope() as session:
            query = session.query(model.NendoCollectionDB)
            if details:
                query = query.options(
                    joinedload(model.NendoCollectionDB.related_tracks).joinedload(
                        model.TrackCollectionRelationshipDB.source,
                    ),
                )
            collection_db = query.filter(
                model.NendoCollectionDB.id == collection_id,
            ).first()

            if details:
                collection = (
                    schema.NendoCollection.model_validate(collection_db)
                    if collection_db is not None
                    else None
                )
            else:
                collection = (
                    schema.NendoCollectionSlim.model_validate(collection_db)
                    if collection_db is not None
                    else None
                )
            return collection

    def get_collections(
        self,
        query: Optional[Query] = None,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: Optional[str] = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        session: Optional[Session] = None,
    ) -> Union[List, Iterator]:
        """Get a list of collections.

        Args:
            query (Optional[Query]): Query object to build from.
            user_id (Union[str, UUID], optional): The user ID to filter for.
            order_by (Optional[str]): Key used for ordering the results.
            order (Optional[str]): Order in which to retrieve results
                ("asc" or "desc").
            limit (Optional[int]): Limit the number of returned results.
            offset (Optional[int]): Offset into the paginated results (requires limit).
            session (sqlalchemy.Session): Session object to commit to.

        Returns:
            Union[List, Iterator]: List or generator of collections, depending on the
                configuration variable stream_mode
        """
        user_id = self._ensure_user_uuid(user_id)
        with self.session_scope() as session:
            query = session.query(model.NendoCollectionDB).filter(
                model.NendoCollectionDB.user_id == user_id,
            )
            return self._get_collections_db(
                query, user_id, order_by, order, limit, offset, session,
            )

    @schema.NendoPlugin.stream_output
    def _get_collections_db(
        self,
        query: Optional[Query] = None,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: Optional[str] = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
        session: Optional[Session] = None,
    ) -> Union[List, Iterator]:
        """Get a list of collections from the DB."""
        user_id = self._ensure_user_uuid(user_id)
        s = session or self.session_scope()
        with s as session_local:
            if query:
                query_local = query
            else:
                query_local = session_local.query(model.NendoCollectionDB).filter(
                    model.NendoCollectionDB.user_id == user_id,
                )

            if order_by:
                if order_by == "random":
                    query_local = query_local.order_by(func.random())
                elif order == "desc":
                    query_local = query_local.order_by(desc(order_by))
                else:
                    query_local = query_local.order_by(asc(order_by))

            if limit:
                query_local = query_local.limit(limit)
                if offset:
                    query_local = query_local.offset(offset)

            if self.config.stream_chunk_size > 1:
                chunk = []
                for collection in query:
                    chunk.append(schema.NendoCollection.model_validate(collection))
                    if len(chunk) == self.config.stream_chunk_size:
                        yield chunk
                        chunk = []
                if chunk:  # yield remaining tracks in non-full chunk
                    yield chunk
            else:
                for collection in query:
                    yield schema.NendoCollection.model_validate(collection)

    def find_collections(
        self,
        value: str = "",
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: Optional[str] = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[List, Iterator]:
        """Find collections with a search term in the description or meta field.

        Args:
            value (str): Term to be searched for in the description and meta field.
            user_id (Union[str, UUID], optional): The user ID to filter for.
            order_by (Optional[str]): Key used for ordering the results.
            order (Optional[str]): Order in which to retrieve results ("asc" or "desc").
            limit (Optional[int]): Limit the number of returned results.
            offset (Optional[int]): Offset into the paginated results (requires limit).

        Returns:
            Union[List, Iterator]: List or generator of collections, depending on the
                configuration variable stream_mode
        """
        user_id = self._ensure_user_uuid(user_id)
        with self.session_scope() as session:
            query = session.query(model.NendoCollectionDB).filter(
                and_(
                    or_(
                        model.NendoCollectionDB.name.ilike(f"%{value}%"),
                        model.NendoCollectionDB.description.ilike(f"%{value}%"),
                        # cast(
                        #     model.NendoCollectionDB.meta, Text()).ilike(f"%{value}%"
                        # ),
                    ),
                    model.NendoCollectionDB.user_id == user_id,
                )
            )
            return self._get_collections_db(
                query, user_id, order_by, order, limit, offset, session,
            )

    def get_related_collections(
        self,
        collection_id: Union[str, uuid.UUID],
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: Optional[str] = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[List, Iterator]:
        """Get collections with a relationship to the collection with collection_id.

        Args:
            collection_id (str): ID of the collection to be searched for.
            user_id (Union[str, UUID], optional): The user ID to filter for.
            order_by (Optional[str]): Key used for ordering the results.
            order (Optional[str]): Order in which to retrieve results ("asc" or "desc").
            limit (Optional[int]): Limit the number of returned results.
            offset (Optional[int]): Offset into the paginated results (requires limit).

        Returns:
            Union[List, Iterator]: List or generator of collections, depending on the
                configuration variable stream_mode
        """
        user_id = self._ensure_user_uuid(user_id)
        with self.session_scope() as session:
            query = self._get_related_collections_query(
                collection_id=ensure_uuid(collection_id),
                session=session,
                user_id=user_id,
            )
            return self._get_collections_db(query, order_by, order, limit, offset, session)

    def remove_track_from_collection(
        self,
        track_id: Union[str, uuid.UUID],
        collection_id: Union[str, uuid.UUID],
    ) -> bool:
        """Deletes a relationship from the track to the collection.

        Args:
            track_id (Union[str, uuid.UUID]): ID
            collection_id (Union[str, uuid.UUID]): Collection id.

        Returns:
            success (bool): True if removal was successful, False otherwise.
        """
        with self.session_scope() as session:
            collection_id = ensure_uuid(collection_id)
            track_id = ensure_uuid(track_id)

            # # Check the collection and track objects
            # collection = (
            #     session.query(model.NendoCollectionDB)
            #     .filter_by(id=collection_id)
            #     .first()
            # )
            # track = session.query(model.NendoTrackDB).filter_by(id=track_id).first()
            # if not collection:
            #     raise ValueError("The collection does not exist")
            # if not track:
            #     raise ValueError("The track does not exist")

            # Remove the relationship from the track to the collection
            return self._remove_track_from_collection_db(
                track_id=track_id, collection_id=collection_id, session=session,
            )

    def update_collection(
        self,
        collection: schema.NendoCollection,
    ) -> schema.NendoCollection:
        """Updates the given collection by storing it to the database.

        Args:
            collection (NendoCollection): The collection to store.

        Raises:
            NendoCollectionNotFoundError: If the collection with
                the given ID was not found.

        Returns:
            NendoCollection: The updated collection.
        """
        with self.session_scope() as session:
            collection_db = (
                session.query(model.NendoCollectionDB)
                .filter_by(id=collection.id)
                .first()
            )

            if collection_db is None:
                raise schema.NendoCollectionNotFoundError(
                    "Collection not found", collection.id,
                )

            collection_db.name = collection.name
            collection_db.user_id = collection.user_id
            collection_db.description = collection.description
            collection_db.collection_type = collection.collection_type
            collection_db.visibility = collection.visibility
            collection_db.meta = collection.meta

            session.commit()
            session.refresh(collection_db)
            return schema.NendoCollection.model_validate(collection_db)

    def remove_collection(
        self,
        collection_id: uuid.UUID,
        remove_relationships: bool = False,
    ) -> bool:
        """Deletes the collection identified by `collection_id`.

        Args:
            collection_id (uuid.UUID): ID of the collection to remove.
            remove_relationships (bool, optional):
                If False prevent deletion if related tracks exist,
                if True delete relationships together with the object.
                Defaults to False.

        Returns:
            bool: True if deletion was successful, False otherwise.
        """
        with self.session_scope() as session:
            # has_related_track = (
            #     session.query(model.TrackCollectionRelationshipDB)
            #     .filter(
            #         model.TrackCollectionRelationshipDB.target_id
            #         == collection_id
            #     )
            #     .first()
            # ) is not False
            has_related_collection = (
                session.query(model.CollectionCollectionRelationshipDB)
                .filter(
                    or_(
                        model.CollectionCollectionRelationshipDB.source_id
                        == collection_id,
                        model.CollectionCollectionRelationshipDB.target_id
                        == collection_id,
                    ),
                )
                .first()
            ) is not None

            if has_related_collection:  # or has_related_track:
                if remove_relationships:
                    session.query(model.CollectionCollectionRelationshipDB).filter(
                        or_(
                            model.CollectionCollectionRelationshipDB.target_id
                            == collection_id,
                            model.CollectionCollectionRelationshipDB.source_id
                            == collection_id,
                        ),
                    ).delete()
                    session.commit()
                else:
                    logger.warning(
                        "Cannot remove due to existing relationships. "
                        "Set `remove_relationships=True` to remove them.",
                    )
                    return False

            # clean up track-collection relationships
            session.query(model.TrackCollectionRelationshipDB).filter(
                model.TrackCollectionRelationshipDB.target_id == collection_id,
            ).delete()
            session.commit()

            # delete actual target
            session.query(model.NendoCollectionDB).filter(
                model.NendoCollectionDB.id == collection_id,
            ).delete()

        return True

    def export_collection(
        self,
        collection_id: Union[str, uuid.UUID],
        export_path: str,
        filename_suffix: str = "nendo",
        file_format: str = "wav",
    ) -> List[str]:
        """Export the collection to a directory.

        Args:
            collection_id (Union[str, uuid.UUID]): The ID of the target
                collection to export.
            export_path (str): Path to a directory into which the collection's tracks
                should be exported.
            filename_suffix (str): The suffix which should be appended to each
                exported track's filename.
            file_format (str, optional): Format of the exported track. Ignored if
                file_path is a full file path. Defaults to "wav".

        Returns:
            List[str]: A list with all full paths to the exported files.
        """
        collection_tracks = self.get_collection_tracks(collection_id)
        now = datetime.now().strftime("%Y%m%d%H%M%S") # noqa: DTZ005
        if not os.path.isdir(export_path):
            logger.info(
                f"Export path {export_path} does not exist, creating now.",
            )
            os.makedirs(export_path, exist_ok=True)
        track_file_paths = []
        for track in collection_tracks:
            if track.has_meta("original_filename"):
                original_filename = track.get_meta("original_filename")
            else:
                original_filename = track.resource.file_name
            file_name = f"{original_filename}_{filename_suffix}_{now}.{file_format}"
            file_path = os.path.join(export_path, file_name)
            track_file_path = self.export_track(
                track_id = track.id,
                file_path = file_path,
                file_format = file_format,
            )
            track_file_paths.append(track_file_path)
        return track_file_paths

    # =========================
    #
    # BLOB MANAGEMENT FUNCTIONS
    #
    # =========================

    def _upsert_blob_db(
        self, blob: schema.NendoBlobBase, session: Session,
    ) -> model.NendoBlobDB:
        """Create blob in DB or update if it exists.

        Args:
            blob (schema.NendoBlobCreate): Blob object to be created
            session (Session): Session to be used for the transaction

        Returns:
            model.NendoBlobDB: The ORM model object of the upserted blob
        """
        if type(blob) == schema.NendoBlobCreate:
            # create new blob
            db_blob = model.NendoBlobDB(**blob.model_dump())
            session.add(db_blob)
        else:
            # update existing blob
            db_blob = (
                session.query(model.NendoBlobDB).filter_by(id=blob.id).one_or_none()
            )
            if db_blob is None:
                raise schema.NendoBlobNotFoundError("Blob not found", target_id=blob.id)
            db_blob.resource = blob.resource.model_dump()
        session.commit()
        return db_blob

    def _create_blob_from_bytes(
        self,
        data: bytes,
        user_id: Optional[uuid.UUID] = None,
    ) -> schema.NendoBlobCreate:
        """Create a blob from the given bytes."""
        target_file = None
        try:
            target_file = self.storage_driver.save_bytes(
                file_name=self.storage_driver.generate_filename(
                    filetype="pkl",
                    user_id=str(user_id) if user_id else str(self.user.id),
                ),
                data=data,
                user_id=str(user_id) if user_id else str(self.user.id),
            )
        except Exception as e:  # noqa: BLE001
            raise schema.NendoLibraryError(
                f"Failed writing file {target_file} to the library. Error: {e}.",
            ) from None
        resource = schema.NendoResource(
            file_path=self.storage_driver.get_file_path(
                src=target_file,
                user_id=str(user_id) if user_id else str(self.user.id),
            ),
            file_name=self.storage_driver.get_file_name(
                src=target_file,
                user_id=str(user_id) if user_id else str(self.user.id),
            ),
            resource_type="blob",
            location=self.storage_driver.get_driver_location(),
            meta={},
        )
        return schema.NendoBlobCreate(resource=resource.model_dump(), user_id=self.user.id)

    def _create_blob_from_file(
        self,
        file_path: FilePath,
        copy_to_library: Optional[bool] = None,
        user_id: Optional[uuid.UUID] = None,
    ) -> schema.NendoBlobCreate:
        """Create a blob from a given filepath."""
        target_file = None
        if not os.path.isfile(file_path):
            raise schema.NendoResourceError("File not found", file_path)

        copy_to_library = copy_to_library or self.config.copy_to_library
        meta = {}
        meta.update({"checksum": md5sum(file_path)})
        if copy_to_library:
            try:
                file_stats = os.stat(file_path)
                meta.update(
                    {
                        "original_filename": os.path.basename(file_path),
                        "original_filepath": os.path.dirname(file_path),
                        "original_size": file_stats.st_size,
                        "original_checksum": md5sum(file_path),
                    },
                )

                target_file = self.storage_driver.save_file(
                    file_name=self.storage_driver.generate_filename(
                        filetype=os.path.splitext(file_path)[1][1:],  # without the dot
                        user_id=user_id,
                    ),
                    file_path=file_path,
                    user_id=str(user_id) if user_id else str(self.user.id),
                )
                location = self.storage_driver.get_driver_location()
            except Exception as e:  # noqa: BLE001
                raise schema.NendoLibraryError(
                    f"Failed storing blob {target_file}. Error: {e}.",
                ) from None
        else:
            target_file = file_path
            location = schema.ResourceLocation.original

        resource = schema.NendoResource(
            file_path=os.path.dirname(target_file),
            file_name=os.path.basename(target_file),
            location=location,
            meta=meta or {},
        )
        return schema.NendoBlobCreate(resource=resource.model_dump(), user_id=self.user.id)

    def load_blob(
        self, blob_id: uuid.UUID, user_id: Optional[Union[str, uuid.UUID]] = None,
    ) -> schema.NendoBlob:
        """Loads a blob of data into memory.

        Args:
            blob_id (uuid.UUID): The UUID of the blob.
            user_id (Optional[Union[str, uuid.UUID]], optional): ID of the user
                who's loading the blob.

        Returns:
            schema.NendoBlob: The loaded blob.
        """
        user_id = self._ensure_user_uuid(user_id)
        with self.session_scope() as session:
            blob_db = (
                session.query(model.NendoBlobDB)
                .filter(model.NendoBlobDB.id == blob_id)
                .one_or_none()
            )
            if blob_db is not None:
                blob = schema.NendoBlob.model_validate(blob_db)
                local_blob = self.storage_driver.as_local(
                    file_path=blob.resource.src,
                    location=blob.resource.location,
                    user_id=user_id,
                )

                # load blob data into memory
                if os.path.splitext(local_blob)[1] == ".pkl":
                    with open(local_blob, "rb") as f:
                        blob.data = pickle.load(f)  # noqa: S301
                elif os.path.splitext(local_blob)[1] == ".npy":
                    blob.data = np.load(local_blob)
                elif os.path.splitext(local_blob)[1] == ".wav":
                    librosa.load(local_blob, mono=False)
                else:
                    logger.error(
                        "Blob file format not recognized. "
                        "Returning blob with empty data.",
                    )

        return blob

    def store_blob(
        self, file_path: Union[str, FilePath], user_id: Optional[Union[str, uuid.UUID]] = None
    ) -> schema.NendoBlob:
        """Stores a blob of data.

        Args:
            file_path (Union[str, FilePath]): The blob to store.
            user_id (Optional[Union[str, uuid.UUID]], optional): ID of the user
                who's storing the file to blob.

        Returns:
            schema.NendoBlob: The stored blob.
        """
        user_id = self._ensure_user_uuid(user_id)
        blob = self._create_blob_from_file(file_path=file_path, user_id=user_id)
        if blob is not None:
            with self.session_scope() as session:
                db_blob = self._upsert_blob_db(blob, session)
                blob = schema.NendoBlob.model_validate(db_blob)
        return blob

    def store_blob_from_bytes(
        self,
        data: bytes,
        user_id: Optional[Union[str, uuid.UUID]] = None,
    ) -> schema.NendoBlob:
        """Stores a data of type `bytes` to a blob.

        Args:
            data (bytes): The blob to store.
            user_id (Optional[Union[str, uuid.UUID]], optional): ID of the user
                who's storing the bytes to blob.

        Returns:
            schema.NendoBlob: The stored blob.
        """
        user_id = self._ensure_user_uuid(user_id)
        blob = self._create_blob_from_bytes(data=data, user_id=user_id)
        if blob is not None:
            with self.session_scope() as session:
                db_blob = self._upsert_blob_db(blob, session)
                blob = schema.NendoBlob.model_validate(db_blob)
        return blob

    def remove_blob(
        self,
        blob_id: uuid.UUID,
        remove_resources: bool = True,
        user_id: Optional[Union[str, uuid.UUID]] = None,
    ) -> bool:
        """Deletes a blob of data.

        Args:
            blob_id (uuid.UUID): The UUID of the blob.
            remove_resources (bool): If True, remove associated resources.
            user_id (Optional[Union[str, uuid.UUID]], optional): ID of the user
                who's removing the blob.

        Returns:
            success (bool): True if removal was successful, False otherwise
        """
        user_id = self._ensure_user_uuid(user_id)
        with self.session_scope() as session:
            target = (
                session.query(model.NendoBlobDB)
                .filter(model.NendoBlobDB.id == blob_id)
                .first()
            )
            session.delete(target)
        if remove_resources:
            logger.info("Removing resources associated with Blob %s", str(blob_id))
            try:
                self.storage_driver.remove_file(
                    file_name=target.resource["file_name"],
                    user_id=user_id,
                )
            except Exception as e:  # noqa: BLE001
                logger.error("Removing %s failed: %s", target.resource.model_dump().src, e)
        return True

    # ==================================
    #
    # MISCELLANEOUS MANAGEMENT FUNCTIONS
    #
    # ==================================

    def reset(
        self, force: bool = False, user_id: Optional[Union[str, uuid.UUID]] = None,
    ) -> None:
        """Reset the nendo library.

        Erase all tracks, collections and relationships. Ask before erasing everything.

        Args:
            force (bool, optional): Flag that specifies whether to ask the user for
                confirmation of the operation. Default is to ask the user.
            user_id (Optional[Union[str, uuid.UUID]], optional): ID of the user
                who's resetting the library. If none is given, the configured
                nendo default user will be used.
        """
        user_id = self._ensure_user_uuid(user_id)
        should_proceed = (
            force
            or input(
                "Are you sure you want to reset the library? "
                "This will purge ALL tracks, collections and relationships!"
                "Enter 'y' to confirm: ",
            ).lower()
            == "y"
        )

        if not should_proceed:
            logger.info("Reset operation cancelled.")
            return

        logger.info("Resetting nendo library.")
        with self.session_scope() as session:
            # delete all relationships
            session.query(model.TrackTrackRelationshipDB).delete()
            session.query(model.TrackCollectionRelationshipDB).delete()
            session.query(model.CollectionCollectionRelationshipDB).delete()
            # delete all plugin data
            session.query(model.NendoPluginDataDB).delete()
            session.commit()
            # delete all collections
            session.query(model.NendoCollectionDB).delete()
            # delete all tracks
            session.query(model.NendoTrackDB).delete()
        # remove files
        for library_file in self.storage_driver.list_files(user_id=user_id):
            self.storage_driver.remove_file(file_name=library_file, user_id=user_id)



# -*- encoding: utf-8 -*-
"""Module implementing the NendoLibraryPlugin using DuckDB.

A lightweight and fast implementation that is used as
Nendo's default library implementation.
"""

import logging
from typing import Any, Optional

from requests import Session
from sqlalchemy import Engine, create_engine

from nendo import schema
from nendo.config import NendoConfig, get_settings
from nendo.library import model
from nendo.utils import play_signal

from .sqlalchemy_library import SqlAlchemyNendoLibrary

logger = logging.getLogger("nendo")


class DuckDBLibrary(SqlAlchemyNendoLibrary):
    """DuckDB-based implementation of the Nendo Library.

    Inherits almost all functions from the SQLAlchemy implementation of the
    `NendoLibraryPlugin` and only differs in the way it connects to the database.
    """

    config: NendoConfig = None
    user: schema.NendoUser = None
    db: Engine = None
    storage_driver: schema.NendoStorage = None

    def __init__(
        self,
        config: Optional[NendoConfig] = None,
        db: Optional[Engine] = None,
        session: Optional[Session] = None,
        **kwargs: Any,
    ) -> None:
        """Configure and connect to the database."""
        super().__init__(**kwargs)
        self.config = config or get_settings()

        if self.storage_driver is None:
            self.storage_driver = schema.NendoStorageLocalFS(
                library_path=self.config.library_path, user_id=self.config.user_id,
            )

        self._connect(db, session)

    def _connect(
        self,
        db: Optional[Engine] = None,
        session: Optional[Session] = None,  # noqa: ARG002
    ) -> None:
        """Open local DuckDB session."""
        self.db = db or create_engine(f"duckdb:///{self.config.library_path}/nendo.db")
        model.Base.metadata.create_all(bind=self.db)
        self.user = self.default_user

    def play(self, track: schema.NendoTrack) -> None:
        """Preview an audio track on mac & linux.

        Args:
            track (NendoTrack): The track to play.
        """
        play_signal(track.signal, track.sr)



# -*- encoding: utf-8 -*-
"""Plugin classes of Nendo Core."""
from __future__ import annotations

import functools
import os
from abc import abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)

from pydantic import ConfigDict, DirectoryPath, FilePath

from nendo.schema.core import (
    NendoBlob,
    NendoCollection,
    NendoCollectionSlim,
    NendoPlugin,
    NendoPluginData,
    NendoStorage,
    NendoTrack,
)
from nendo.schema.exception import NendoError, NendoPluginRuntimeError
from nendo.utils import ensure_uuid

if TYPE_CHECKING:
    import uuid

    import numpy as np


class NendoAnalysisPlugin(NendoPlugin):
    """Basic class for nendo analysis plugins.

    Analysis plugins are plugins that analyze a track or a collection
    and add metadata and other properties to the track or collection.
    Decorate your methods with `@NendoAnalysisPlugin.plugin_data` to add the
    return values of your methods as plugin data to the track or collection.

    Decorate your methods with `@NendoAnalysisPlugin.run_track` to run your method
    on a track and use `@NendoAnalysisPlugin.run_collection` to run your method on
    a collection.

    Examples:
        ```python
        from nendo import Nendo, NendoConfig

        class MyPlugin(NendoAnalysisPlugin):
            ...

            @NendoAnalysisPlugin.plugin_data
            def my_plugin_data_function_one(self, track):
                # do something analysis on the track
                return {"key": "value"}

            @NendoAnalysisPlugin.plugin_data
            def my_plugin_data_function_two(self, track):
                # do some more analysis on the track
                return {"key": "value"}

            @NendoAnalysisPlugin.run_track
            def my_run_track_function(self, track):
                my_plugin_data_function_one(track)
                my_plugin_data_function_two(track)
        ```
    """

    # decorators
    # ----------
    @staticmethod
    def plugin_data(
        func: Callable[[NendoPlugin, NendoTrack], Dict[str, Any]],
    ) -> Callable[[NendoPlugin, NendoTrack], Dict[str, Any]]:
        """Decorator to enrich a NendoTrack with data from a plugin.

        Args:
            func: Callable[[NendoPlugin, NendoTrack], Dict[str, Any]]: The function to register.

        Returns:
            Callable[[NendoPlugin, NendoTrack], Dict[str, Any]]: The wrapped function.
        """

        def wrapper(self, track: NendoTrack):
            try:
                f_result = func(self, track)
            except NendoError as e:
                raise NendoPluginRuntimeError(
                    f"Error running plugin function: {e}",
                ) from None
            for k, v in f_result.items():
                track.add_plugin_data(
                    plugin_name=self.plugin_name,
                    plugin_version=self.plugin_version,
                    key=str(k),
                    value=v,
                    # replace=False,
                )
            return f_result

        return wrapper

    # ----------

    @property
    def plugin_type(self) -> str:
        """Return type of plugin."""
        return "AnalysisPlugin"

    @staticmethod
    def run_collection(
        func: Callable[[NendoPlugin, NendoCollection, Any], None],
    ) -> Callable[[NendoPlugin, Any], NendoCollection]:
        """Decorator to register a function as a collection running function for a `NendoAnalysisPlugin`.

        This decorator wraps the function and allows a plugin user to call the plugin with either a collection or a track.

        Args:
            func: Callable[[NendoPlugin, NendoCollection, Any], None]: The function to register.

        Returns:
            Callable[[NendoPlugin, Any], NendoCollection]: The wrapped function.
        """

        @functools.wraps(func)
        def wrapper(self, **kwargs: Any) -> NendoCollection:
            track_or_collection, kwargs = self._get_track_or_collection_from_args(
                **kwargs,
            )
            if isinstance(track_or_collection, NendoCollection):
                func(self, track_or_collection, **kwargs)
                return self.nendo_instance.library.get_collection(
                    track_or_collection.id,
                )

            tmp_collection = self.nendo_instance.library.add_collection(
                name="tmp",
                track_ids=[track_or_collection.id],
                collection_type="temp",
            )
            func(self, tmp_collection, **kwargs)
            return self.nendo_instance.library.get_track(track_or_collection.id)

        return wrapper

    @staticmethod
    def run_track(
        func: Callable[[NendoPlugin, NendoTrack, Any], None],
    ) -> Callable[[NendoPlugin, Any], Union[NendoTrack, NendoCollection]]:
        """Decorator to register a function as a track running function for a `NendoAnalysisPlugin`.

        This decorator wraps the function and allows a plugin user to call the plugin with either a collection or a track.

        Args:
            func: Callable[[NendoPlugin, NendoTrack, Any], None]: The function to register.

        Returns:
            Callable[[NendoPlugin, Any], NendoCollection]: The wrapped function.
        """

        @functools.wraps(func)
        def wrapper(self, **kwargs: Any) -> Union[NendoTrack, NendoCollection]:
            track_or_collection, kwargs = self._get_track_or_collection_from_args(
                **kwargs,
            )
            if isinstance(track_or_collection, NendoTrack):
                func(self, track_or_collection, **kwargs)
                return self.nendo_instance.library.get_track(track_or_collection.id)
            [func(self, track, **kwargs) for track in track_or_collection.tracks()]
            return self.nendo_instance.library.get_collection(track_or_collection.id)

        return wrapper


class NendoGeneratePlugin(NendoPlugin):
    """Basic class for nendo generate plugins.

    Generate plugins are plugins that generate new tracks or collections, either from scratch or
    based on existing tracks or collections.
    Decorate your methods with `@NendoGeneratePlugin.run_track` to run your method
    on a track, use `@NendoGeneratePlugin.run_collection` to run your method on
    a collection and use `@NendoGeneratePlugin.run_signal` to run your method on
    a signal.

    Examples:
        ```python
        from nendo import Nendo, NendoConfig

        class MyPlugin(NendoGeneratePlugin):
            ...

            @NendoAnalysisPlugin.run_track
            def my_generate_track_function(self, track, arg_one="foo"):
                # generate some new audio

                # add audio to the nendo library
                return self.nendo_instance.library.add_related_track_from_signal(
                    signal,
                    sr,
                    related_track_id=track.id,
                )
        ```
    """

    @property
    def plugin_type(self) -> str:
        """Return type of plugin."""
        return "GeneratePlugin"

    @staticmethod
    def run_collection(
        func: Callable[[NendoPlugin, Optional[NendoCollection], Any], NendoCollection],
    ) -> Callable[[NendoPlugin, Any], NendoCollection]:
        """Decorator to register a function as a collection running function for a `NendoGeneratePlugin`.

        This decorator wraps the function and allows a plugin user to call the plugin with either a collection or a track.

        Args:
            func: Callable[[NendoPlugin, NendoCollection, Any], NendoCollection]: The function to register.

        Returns:
            Callable[[NendoPlugin, Any], NendoCollection]: The wrapped function.
        """

        @functools.wraps(func)
        def wrapper(self, **kwargs: Any) -> NendoCollection:
            track_or_collection, kwargs = self._get_track_or_collection_from_args(
                **kwargs,
            )
            if track_or_collection is None:
                return func(self, **kwargs)

            if isinstance(track_or_collection, NendoCollection):
                return func(self, track_or_collection, **kwargs)

            tmp_collection = self.nendo_instance.library.add_collection(
                name="tmp",
                track_ids=[track_or_collection.id],
                collection_type="temp",
            )
            return func(self, tmp_collection, **kwargs)

        return wrapper

    @staticmethod
    def run_signal(
        func: Callable[
            [NendoPlugin, Optional[np.ndarray], Optional[int], Any],
            Tuple[np.ndarray, int],
        ],
    ) -> Callable[[NendoPlugin, Any], Union[NendoTrack, NendoCollection]]:
        """Decorator to register a function as a signal running function for a `NendoGeneratePlugin`.

        This decorator wraps the function and allows a plugin user to call the plugin with either a collection or a track.

        Args:
            func: Callable[[NendoPlugin, np.ndarray, int, Any], Tuple[np.ndarray, int]]: The function to register.

        Returns:
            Callable[[NendoPlugin, Any], NendoCollection]: The wrapped function.
        """

        @functools.wraps(func)
        def wrapper(self, **kwargs: Any) -> Union[NendoTrack, NendoCollection]:
            track_or_collection, kwargs = self._get_track_or_collection_from_args(
                **kwargs,
            )
            if track_or_collection is None:
                signal, sr = func(self, **kwargs)
                return self.nendo_instance.library.add_track_from_signal(
                    signal,
                    sr,
                )
            if isinstance(track_or_collection, NendoTrack):
                signal, sr = track_or_collection.signal, track_or_collection.sr
                new_signal, new_sr = func(self, signal, sr, **kwargs)
                return self.nendo_instance.library.add_related_track_from_signal(
                    new_signal,
                    sr,
                    related_track_id=track_or_collection.id,
                )
            processed_tracks = []
            for track in track_or_collection.tracks():
                new_signal, new_sr = func(
                    self,
                    track.signal,
                    track.sr,
                    **kwargs,
                )
                processed_tracks.append(
                    self.nendo_instance.library.add_related_track_from_signal(
                        new_signal,
                        track.sr,
                        related_track_id=track.id,
                    ),
                )
            return self.nendo_instance.library.add_collection(
                name="tmp",
                track_ids=[track.id for track in processed_tracks],
                collection_type="temp",
            )

        return wrapper

    @staticmethod
    def run_track(
        func: Callable[
            [NendoPlugin, Optional[NendoTrack], Any],
            Union[NendoTrack, List[NendoTrack]],
        ],
    ) -> Callable[[NendoPlugin, Any], Union[NendoTrack, NendoCollection]]:
        """Decorator to register a function as a track running function for a `NendoGeneratePlugin`.

        This decorator wraps the function and allows a plugin user to call the plugin with either a collection or a track.

        Args:
            func: Callable[[NendoPlugin, NendoTrack, Any], Union[NendoTrack, List[NendoTrack]]]: The function to register.

        Returns:
            Callable[[NendoPlugin, Any], NendoCollection]: The wrapped function.
        """

        @functools.wraps(func)
        def wrapper(self, **kwargs: Any) -> Union[NendoTrack, NendoCollection]:
            track_or_collection, kwargs = self._get_track_or_collection_from_args(
                **kwargs,
            )
            processed_tracks = []
            if track_or_collection is None:
                track = func(self, **kwargs)

                # may be multiple tracks as a result
                if not isinstance(track, list):
                    return track
                processed_tracks.extend(track)
            elif isinstance(track_or_collection, NendoTrack):
                track = func(self, track_or_collection, **kwargs)

                # may be multiple tracks as a result
                if not isinstance(track, list):
                    return track
                processed_tracks.extend(track)
            else:
                for track in track_or_collection.tracks():
                    processed_track = func(self, track, **kwargs)

                    # may be multiple tracks as a result
                    if isinstance(processed_track, list):
                        processed_tracks.extend(processed_track)
                    else:
                        processed_tracks.append(processed_track)

            return self.nendo_instance.library.add_collection(
                name="tmp",
                track_ids=[track.id for track in processed_tracks],
                collection_type="temp",
            )

        return wrapper


class NendoEffectPlugin(NendoPlugin):
    """Basic class for nendo effects plugins.

    Effects plugins are plugins that add effects to tracks or collections.
    Decorate your methods with `@NendoGeneratePlugin.run_track` to run your method
    on a track, use `@NendoGeneratePlugin.run_collection` to run your method on
    a collection and use `@NendoGeneratePlugin.run_signal` to run your method on
    a signal.

    Examples:
        ```python
        from nendo import Nendo, NendoConfig

        class MyPlugin(NendoGeneratePlugin):
            ...

            @NendoAnalysisPlugin.run_signal
            def my_effect_function(self, signal, sr, arg_one="foo"):
                # add some effect to the signal
                new_signal = apply_effect(signal, sr, arg_one)

                return new_signal, sr
        ```
    """

    @property
    def plugin_type(self) -> str:
        """Return type of plugin."""
        return "EffectPlugin"

    @staticmethod
    def run_collection(
        func: Callable[[NendoPlugin, NendoCollection, Any], NendoCollection],
    ) -> Callable[[NendoPlugin, Any], NendoCollection]:
        """Decorator to register a function as a collection running function for a `NendoEffectPlugin`.

        This decorator wraps the function and allows a plugin user to call the plugin with either a collection or a track.

        Args:
            func: Callable[[NendoPlugin, NendoCollection, Any], NendoCollection]: The function to register.

        Returns:
            Callable[[NendoPlugin, Any], NendoCollection]: The wrapped function.
        """

        @functools.wraps(func)
        def wrapper(self, **kwargs: Any) -> NendoCollection:
            track_or_collection, kwargs = self._get_track_or_collection_from_args(
                **kwargs,
            )
            if isinstance(track_or_collection, NendoCollection):
                return func(self, track_or_collection, **kwargs)

            tmp_collection = self.nendo_instance.library.add_collection(
                name="tmp",
                track_ids=[track_or_collection.id],
                collection_type="temp",
            )
            return func(self, tmp_collection, **kwargs)

        return wrapper

    @staticmethod
    def run_signal(
        func: Callable[[NendoPlugin, np.ndarray, int, Any], Tuple[np.ndarray, int]],
    ) -> Callable[[NendoPlugin, Any], Union[NendoTrack, NendoCollection]]:
        """Decorator to register a function as a signal running function for a `NendoEffectPlugin`.

        This decorator wraps the function and allows a plugin user to call the plugin with either a collection or a track.

        Args:
            func: Callable[[NendoPlugin, np.ndarray, int, Any], Tuple[np.ndarray, int]]: The function to register.

        Returns:
            Callable[[NendoPlugin, Any], NendoTrack]: The wrapped function.
        """

        @functools.wraps(func)
        def wrapper(self, **kwargs: Any) -> Union[NendoTrack, NendoCollection]:
            track_or_collection, kwargs = self._get_track_or_collection_from_args(
                **kwargs,
            )
            processed_tracks = []
            if isinstance(track_or_collection, NendoTrack):
                signal, sr = track_or_collection.signal, track_or_collection.sr
                new_signal, new_sr = func(self, signal, sr, **kwargs)

                # TODO update track instead of adding a new one to the library
                return self.nendo_instance.library.add_related_track_from_signal(
                    new_signal,
                    sr,
                    related_track_id=track_or_collection.id,
                )

            for track in track_or_collection.tracks():
                new_signal, new_sr = func(
                    self,
                    track.signal,
                    track.sr,
                    **kwargs,
                )

                # TODO update track instead of adding a new one to the library
                processed_tracks.append(
                    self.nendo_instance.library.add_related_track_from_signal(
                        new_signal,
                        track.sr,
                        related_track_id=track.id,
                    ),
                )

            return self.nendo_instance.library.add_collection(
                name="tmp",
                track_ids=[track.id for track in processed_tracks],
                collection_type="temp",
            )

        return wrapper

    @staticmethod
    def run_track(
        func: Callable[
            [NendoPlugin, NendoTrack, Any],
            Union[NendoTrack, List[NendoTrack]],
        ],
    ) -> Callable[[NendoPlugin, Any], NendoTrack]:
        """Decorator to register a function as a track running function for a `NendoEffectPlugin`.

        This decorator wraps the function and allows a plugin user to call the plugin with either a collection or a track.

        Args:
            func: Callable[[NendoPlugin, NendoTrack, Any], Union[NendoTrack, List[NendoTrack]]]: The function to register.

        Returns:
            Callable[[NendoPlugin, Any], NendoTrack]: The wrapped function.
        """

        @functools.wraps(func)
        def wrapper(self, **kwargs: Any) -> Union[NendoTrack, NendoCollection]:
            track_or_collection, kwargs = self._get_track_or_collection_from_args(
                **kwargs,
            )
            if isinstance(track_or_collection, NendoTrack):
                return func(self, track_or_collection, **kwargs)

            processed_tracks = [
                func(self, track, **kwargs) for track in track_or_collection.tracks()
            ]
            return self.nendo_instance.library.add_collection(
                name="tmp",
                track_ids=[track.id for track in processed_tracks],
                collection_type="temp",
            )

        return wrapper


class NendoLibraryPlugin(NendoPlugin):
    """Basic class for nendo library plugins."""

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    storage_driver: Optional[NendoStorage] = None

    # ==========================
    #
    # TRACK MANAGEMENT FUNCTIONS
    #
    # ==========================

    @abstractmethod
    def add_track(
        self,
        file_path: Union[FilePath, str],
        track_type: str = "track",
        copy_to_library: Optional[bool] = None,
        skip_duplicate: Optional[bool] = None,
        user_id: Optional[uuid.UUID] = None,
        meta: Optional[dict] = None,
    ) -> NendoTrack:
        """Add the track given by path to the library.

        Args:
            file_path (Union[FilePath, str]): Path to the file to add as track.
            track_type (str): Type of the track. Defaults to "track".
            copy_to_library (bool, optional): Flag that specifies whether
                the file should be copied into the library directory.
                Defaults to None.
            skip_duplicate (bool, optional): Flag that specifies whether a
                file should be added that already exists in the library, based on its
                file checksum. Defaults to None.
            user_id (UUID, optional): ID of user adding the track.
            meta (dict, optional): Metadata to attach to the track upon adding.

        Returns:
            NendoTrack: The track that was added to the library.
        """
        raise NotImplementedError

    @abstractmethod
    def add_track_from_signal(
        self,
        signal: np.ndarray,
        sr: int,
        track_type: str = "track",
        user_id: Optional[uuid.UUID] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> NendoTrack:
        """Add a track to the library that is described by the given signal.

        Args:
            signal (np.ndarray): The numpy array containing the audio signal.
            sr (int): Sample rate
            track_type (str): Track type. Defaults to "track".
            user_id (UUID, optional): The ID of the user adding the track.
            meta (Dict[str, Any], optional): Track metadata. Defaults to {}.

        Returns:
            schema.NendoTrack: The added NendoTrack
        """
        raise NotImplementedError

    @abstractmethod
    def add_related_track(
        self,
        file_path: Union[FilePath, str],
        related_track_id: Union[str, uuid.UUID],
        track_type: str = "str",
        user_id: Optional[Union[str, uuid.UUID]] = None,
        track_meta: Optional[Dict[str, Any]] = None,
        relationship_type: str = "relationship",
        meta: Optional[Dict[str, Any]] = None,
    ) -> NendoTrack:
        """Add track that is related to another `NendoTrack`.

        Add the track found in the given path to the library and create a relationship
        in the new track that points to the track identified by related_to.

        Args:
            file_path (Union[FilePath, str]): Path to the file to add as track.
            related_track_id (Union[str, uuid.UUID]): ID of the related track.
            track_type (str): Track type. Defaults to "track".
            user_id (Union[str, UUID], optional): ID of the user adding the track.
            track_meta (dict, optional): Dictionary containing the track metadata.
            relationship_type (str): Type of the relationship.
                Defaults to "relationship".
            meta (dict): Dictionary containing metadata about
                the relationship. Defaults to {}.

        Returns:
            NendoTrack: The track that was added to the Library
        """
        raise NotImplementedError

    @abstractmethod
    def add_related_track_from_signal(
        self,
        signal: np.ndarray,
        sr: int,
        related_track_id: Union[str, uuid.UUID],
        track_type: str = "track",
        user_id: Optional[uuid.UUID] = None,
        track_meta: Optional[Dict[str, Any]] = None,
        relationship_type: str = "relationship",
        meta: Optional[Dict[str, Any]] = None,
    ) -> NendoTrack:
        """Add signal as track that is related to another `NendoTrack`.

        Add the track represented by the provided signal to the library and create a
        relationship in the new track that points to the track passed as related_to.

        Args:
            signal (np.ndarray): Waveform of the track in numpy array form.
            sr (int): Sampling rate of the waveform.
            related_track_id (str | uuid.UUID): ID to which the relationship
                should point to.
            track_type (str): Track type. Defaults to "track".
            user_id (UUID, optional): ID of the user adding the track.
            track_meta  (dict, optional): Dictionary containing the track metadata.
            relationship_type (str): Type of the relationship.
                Defaults to "relationship".
            meta (dict): Dictionary containing metadata about
                the relationship. Defaults to {}.

        Returns:
            NendoTrack: The added track with the relationship.
        """
        raise NotImplementedError

    @abstractmethod
    def add_tracks(
        self,
        path: Union[DirectoryPath, str],
        track_type: str = "track",
        user_id: Optional[Union[str, uuid.UUID]] = None,
        copy_to_library: Optional[bool] = None,
        skip_duplicate: bool = True,
    ) -> NendoCollection:
        """Scan the provided path and upsert the information into the library.

        Args:
            path (Union[DirectoryPath, str]): Path to the directory to scan.
            track_type (str): Track type. Defaults to "track".
            user_id (UUID, optional): The ID of the user adding the tracks.
            copy_to_library (Optional[bool], optional): Flag that specifies whether
                the file should be copied into the library directory.
                Defaults to None.
            skip_duplicate (Optional[bool], optional): Flag that specifies whether a
                file should be added that already exists in the library, based on its
                file checksum. Defaults to None.

        Returns:
            collection (NendoCollection): The collection of tracks that were added to the Library
        """
        raise NotImplementedError

    @abstractmethod
    def update_track(
        self,
        track: NendoTrack,
    ) -> NendoTrack:
        """Updates the given collection by storing it to the database.

        Args:
            track (NendoTrack): The track to be stored to the database.

        Raises:
            NendoTrackNotFoundError: If the track passed to the function
                does not exist in the database.

        Returns:
            NendoTrack: The updated track.
        """
        raise NotImplementedError

    @abstractmethod
    def add_plugin_data(
        self,
        track_id: Union[str, uuid.UUID],
        plugin_name: str,
        plugin_version: str,
        key: str,
        value: str,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        replace: bool = False,
    ) -> NendoPluginData:
        """Add plugin data to a NendoTrack and persist changes into the DB.

        Args:
            track_id (Union[str, uuid.UUID]): ID of the track to which
                the plugin data should be added.
            plugin_name (str): Name of the plugin.
            plugin_version (str): Version of the plugin.
            key (str): Key under which to save the data.
            value (str): Data to  save.
            user_id (uuid4, optional): ID of user adding the plugin data.
            replace (bool, optional): Flag that determines whether
                the last existing data point for the given plugin name and -version
                is overwritten or not. Defaults to False.

        Returns:
            NendoPluginData: The saved plugin data as a NendoPluginData object.
        """

    @abstractmethod
    def get_track(self, track_id: Any) -> NendoTrack:
        """Get a single track from the library by ID.

        If no track with the given ID was found, return None.

        Args:
            track_id (Any): The ID of the track to get

        Returns:
            track (NendoTrack): The track with the given ID
        """
        raise NotImplementedError

    @abstractmethod
    @NendoPlugin.stream_output
    def get_tracks(
        self,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: str = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[List, Iterator]:
        """Get tracks based on the given query parameters.

        Args:
            user_id (Union[str, UUID], optional): ID of user getting the tracks.
            order_by (Optional[str]): Key used for ordering the results.
            order (Optional[str]): Order in which to retrieve results ("asc" or "desc").
            limit (Optional[int]): Limit the number of returned results.
            offset (Optional[int]): Offset into the paginated results (requires limit).

        Returns:
            Union[List, Iterator]: List or generator of tracks, depending on the
                configuration variable stream_mode
        """
        raise NotImplementedError

    @abstractmethod
    def get_related_tracks(
        self,
        track_id: Union[str, uuid.UUID],
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: Optional[str] = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[List, Iterator]:
        """Get tracks with a relationship to the track with track_id.

        Args:
            track_id (str): ID of the track to be searched for.
            user_id (Union[str, UUID], optional): The user ID to filter for.
            order_by (Optional[str]): Key used for ordering the results.
            order (Optional[str]): Order in which to retrieve results ("asc" or "desc").
            limit (Optional[int]): Limit the number of returned results.
            offset (Optional[int]): Offset into the paginated results (requires limit).

        Returns:
            Union[List, Iterator]: List or generator of tracks, depending on the
                configuration variable stream_mode
        """
        raise NotImplementedError

    @abstractmethod
    def find_tracks(
        self,
        value: str,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[List, Iterator]:
        """Find tracks by searching for a string through the resource metadata.

        Args:
            value (str): The search value to filter by.
            user_id (Union[str, UUID], optional): The user ID to filter for.
            order_by (str, optional): Ordering.
            limit (str, optional): Pagination limit.
            offset (str, optional): Pagination offset.

        Returns:
            Union[List, Iterator]: List or generator of tracks, depending on the
                configuration variable stream_mode
        """
        raise NotImplementedError

    @abstractmethod
    def filter_tracks(
        self,
        filters: Optional[dict] = None,
        resource_filters: Optional[Dict[str, Any]] = None,
        track_type: Optional[Union[str, List[str]]] = None,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        collection_id: Optional[Union[str, uuid.UUID]] = None,
        plugin_names: Optional[List[str]] = None,
        order_by: Optional[str] = None,
        order: str = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[List, Iterator]:
        """Obtain tracks from the db by filtering over plugin data.

        Args:
            filters (Optional[dict]): Dictionary containing the filters to apply.
                Defaults to None.
            resource_filters (dict): Dictionary containing the keywords to search for
                over the track.resource.meta field. The dictionary's values
                should contain singular search tokens and  the keys currently have no
                effect but might in the future. Defaults to {}.
            track_type (Union[str, List[str]], optional): Track type to filter for.
                Can be a singular type or a list of types. Defaults to None.
            user_id (Union[str, UUID], optional): The user ID to filter for.
            collection_id (Union[str, uuid.UUID], optional): Collection id to
                which the filtered tracks must have a relationship. Defaults to None.
            plugin_names (list, optional): List used for applying the filter only to
                data of certain plugins. If None, all plugin data related to the track
                is used for filtering.
            order_by (str, optional): Key used for ordering the results.
            order (str, optional): Ordering ("asc" vs "desc"). Defaults to "asc".
            limit (int, optional): Limit the number of returned results.
            offset (int, optional): Offset into the paginated results (requires limit).

        Returns:
            Union[List, Iterator]: List or generator of tracks, depending on the
                configuration variable stream_mode
        """
        raise NotImplementedError

    @abstractmethod
    def remove_track(
        self,
        track_id: Union[str, uuid.UUID],
        remove_relationships: bool = False,
        remove_plugin_data: bool = False,
        remove_resources: bool = True,
        user_id: Optional[Union[str, uuid.UUID]] = None,
    ) -> bool:
        """Delete track from library by ID.

        Args:
            track_id (Union[str, uuid.UUID]): The ID of the track to remove.
            remove_relationships (bool):
                If False prevent deletion if related tracks exist,
                if True delete relationships together with the object.
            remove_plugin_data (bool):
                If False prevent deletion if related plugin data exist,
                if True delete plugin data together with the object.
            remove_resources (bool):
                If False, keep the related resources, e.g. files,
                if True, delete the related resources.
            user_id (Union[str, UUID], optional): The ID of the user
                owning the track.

        Returns:
            success (bool): True if removal was successful, False otherwise
        """
        raise NotImplementedError

    def export_track(
        self,
        track_id: Union[str, uuid.UUID],
        file_path: str,
        file_format: str = "wav",
    ) -> str:
        """Export the track to a file.

        Args:
            track_id (Union[str, uuid.UUID]): The ID of the target track to export.
            file_path (str): Path to the exported file. Can be either a full
                file path or a directory path. If a directory path is given,
                a filename will be automatically generated and the file will be
                exported to the format specified as file_format. If a full file
                path is given, the format will be deduced from the path and the
                file_format parameter will be ignored.
            file_format (str, optional): Format of the exported track. Ignored if
                file_path is a full file path. Defaults to "wav".

        Returns:
            str: The path to the exported file.
        """
        raise NotImplementedError

    # ===============================
    #
    # COLLECTION MANAGEMENT FUNCTIONS
    #
    # ===============================

    @abstractmethod
    def add_collection(
        self,
        name: str,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        track_ids: Optional[List[Union[str, uuid.UUID]]] = None,
        description: str = "",
        collection_type: str = "collection",
        meta: Optional[Dict[str, Any]] = None,
    ) -> NendoCollection:
        """Creates a new collection and saves it into the DB.

        Args:
            track_ids (List[Union[str, uuid.UUID]]): List of track ids
                to be added to the collection.
            name (str): Name of the collection.
            user_id (UUID, optional): The ID of the user adding the collection.
            description (str): Description of the collection.
            collection_type (str): Type of the collection. Defaults to "collection".
            meta (Dict[str, Any]): Metadata of the collection.

        Returns:
            schema.NendoCollection: The newly created NendoCollection object.
        """
        raise NotImplementedError

    @abstractmethod
    def add_related_collection(
        self,
        track_ids: List[Union[str, uuid.UUID]],
        collection_id: Union[str, uuid.UUID],
        name: str,
        description: str = "",
        user_id: Optional[Union[str, uuid.UUID]] = None,
        relationship_type: str = "relationship",
        meta: Optional[Dict[str, Any]] = None,
    ) -> NendoCollection:
        """Add a collection that is related to another `NendoCollection`.

        Add a new collection with a relationship to and from the collection
        with the given collection_id.

        Args:
            track_ids (List[Union[str, uuid.UUID]]): List of track ids.
            collection_id (Union[str, uuid.UUID]): Existing collection id.
            name (str): Name of the new related collection.
            description (str): Description of the new related collection.
            user_id (UUID, optional): The ID of the user adding the collection.
            relationship_type (str): Type of the relationship.
            meta (Dict[str, Any]): Meta of the new related collection.

        Returns:
            schema.NendoCollection: The newly added NendoCollection object.
        """
        raise NotImplementedError

    @abstractmethod
    def add_track_to_collection(
        self,
        track_id: Union[str, uuid.UUID],
        collection_id: Union[str, uuid.UUID],
        position: Optional[int] = None,
        meta: Optional[Dict[str, Any]] = None,
    ) -> NendoCollection:
        """Creates a relationship from the track to the collection.

        Args:
            collection_id (Union[str, uuid.UUID]): Collection id.
            track_id (Union[str, uuid.UUID]): Track id.

        Returns:
            schema.NendoCollection: The updated NendoCollection object.
        """
        raise NotImplementedError

    @abstractmethod
    def get_collection_tracks(
        self,
        collection_id: Union[str, uuid.UUID],
    ) -> List[NendoTrack]:
        """Get all tracks of a collection.

        Args:
            collection_id (Union[str, uuid.UUID]): Collection id.

        Returns:
            List[schema.NendoTrack]: List of tracks in the collection.
        """
        raise NotImplementedError

    @abstractmethod
    def get_collection(
        self,
        collection_id: uuid.uuid4,
        details: bool = True,
    ) -> Union[NendoCollection, NendoCollectionSlim]:
        """Get a collection by its ID.

        Args:
            collection_id (uuid.uuid4): ID of the target collection.
            details (bool, optional): Flag that defines whether the result should
                contain all fields or only a Defaults to True.

        Returns:
            Union[NendoCollection, NendoCollectionSlim]: Collection object, compact
                version if the `details` flag has been set to False.
        """
        raise NotImplementedError

    @abstractmethod
    @NendoPlugin.stream_output
    def get_collections(
        self,
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: Optional[str] = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[List, Iterator]:
        """Get a list of collections.

        Args:
            user_id (Union[str, UUID], optional): The user ID to filter for.
            order_by (Optional[str]): Key used for ordering the results.
            order (Optional[str]): Order in which to retrieve results ("asc" or "desc").
            limit (Optional[int]): Limit the number of returned results.
            offset (Optional[int]): Offset into the paginated results (requires limit).

        Returns:
            Union[List, Iterator]: List or generator of collections, depending on the
                configuration variable stream_mode
        """
        raise NotImplementedError

    @abstractmethod
    def find_collections(
        self,
        value: str = "",
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: Optional[str] = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[List, Iterator]:
        """Find collections with a search term in the description or meta field.

        Args:
            value (str): Term to be searched for in the description and meta field.
            user_id (Union[str, UUID], optional): The user ID to filter for.
            order_by (Optional[str]): Key used for ordering the results.
            order (Optional[str]): Order in which to retrieve results ("asc" or "desc").
            limit (Optional[int]): Limit the number of returned results.
            offset (Optional[int]): Offset into the paginated results (requires limit).

        Returns:
            Union[List, Iterator]: List or generator of collections, depending on the
                configuration variable stream_mode
        """
        raise NotImplementedError

    @abstractmethod
    def get_related_collections(
        self,
        collection_id: Union[str, uuid.UUID],
        user_id: Optional[Union[str, uuid.UUID]] = None,
        order_by: Optional[str] = None,
        order: Optional[str] = "asc",
        limit: Optional[int] = None,
        offset: Optional[int] = None,
    ) -> Union[List, Iterator]:
        """Get collections with a relationship to the collection with collection_id.

        Args:
            collection_id (str): ID of the collection to be searched for.
            user_id (Union[str, UUID], optional): The user ID to filter for.
            order_by (Optional[str]): Key used for ordering the results.
            order (Optional[str]): Order in which to retrieve results ("asc" or "desc").
            limit (Optional[int]): Limit the number of returned results.
            offset (Optional[int]): Offset into the paginated results (requires limit).

        Returns:
            Union[List, Iterator]: List or generator of collections, depending on the
                configuration variable stream_mode
        """
        raise NotImplementedError

    @abstractmethod
    def update_collection(
        self,
        collection: NendoCollection,
    ) -> NendoCollection:
        """Updates the given collection by storing it to the database.

        Args:
            collection (NendoCollection): The collection to store.

        Raises:
            NendoCollectionNotFoundError: If the collection with
                the given ID was not found.

        Returns:
            NendoCollection: The updated collection.
        """
        raise NotImplementedError

    @abstractmethod
    def remove_track_from_collection(
        self,
        track_id: Union[str, uuid.UUID],
        collection_id: Union[str, uuid.UUID],
    ) -> bool:
        """Deletes a relationship from the track to the collection.

        Args:
            collection_id (Union[str, uuid.UUID]): Collection id.
            track_id (Union[str, uuid.UUID]): Track id.

        Returns:
            success (bool): True if removal was successful, False otherwise.
        """
        raise NotImplementedError

    @abstractmethod
    def remove_collection(
        self,
        collection_id: uuid.UUID,
        remove_relationships: bool = False,
    ) -> bool:
        """Deletes the collection identified by `collection_id`.

        Args:
            collection_id (uuid.UUID): ID of the collection to remove.
            remove_relationships (bool, optional):
                If False prevent deletion if related tracks exist,
                if True delete relationships together with the object.
                Defaults to False.

        Returns:
            bool: True if deletion was successful, False otherwise.
        """
        raise NotImplementedError

    def export_collection(
        self,
        collection_id: Union[str, uuid.UUID],
        export_path: str,
        filename_suffix: str = "_nendo",
        file_format: str = "wav",
    ) -> List[str]:
        """Export the track to a file.

        Args:
            collection_id (Union[str, uuid.UUID]): The ID of the target
                collection to export.
            export_path (str): Path to a directory into which the collection's tracks
                should be exported.
            filename_suffix (str): The suffix which should be appended to each
                exported track's filename.
            file_format (str, optional): Format of the exported track. Ignored if
                file_path is a full file path. Defaults to "wav".

        Returns:
            List[str]: A list with all full paths to the exported files.
        """
        raise NotImplementedError

    # =========================
    #
    # BLOB MANAGEMENT FUNCTIONS
    #
    # =========================

    @abstractmethod
    def store_blob(
        self,
        file_path: Union[FilePath, str],
        user_id: Optional[Union[str, uuid.UUID]] = None,
    ) -> NendoBlob:
        """Stores a blob of data.

        Args:
            file_path (Union[FilePath, str]): Path to the file to store as blob.
            user_id (Optional[Union[str, uuid.UUID]], optional): ID of the user
                who's storing the file to blob.

        Returns:
            schema.NendoBlob: The stored blob.
        """
        raise NotImplementedError

    @abstractmethod
    def store_blob_from_bytes(
        self,
        data: bytes,
        user_id: Optional[Union[str, uuid.UUID]] = None,
    ) -> NendoBlob:
        """Stores a data of type `bytes` to a blob.

        Args:
            data (bytes): The blob to store.
            user_id (Optional[Union[str, uuid.UUID]], optional): ID of the user
                who's storing the bytes to blob.

        Returns:
            schema.NendoBlob: The stored blob.
        """
        raise NotImplementedError

    @abstractmethod
    def load_blob(
        self,
        blob_id: uuid.UUID,
        user_id: Optional[Union[str, uuid.UUID]] = None,
    ) -> NendoBlob:
        """Loads a blob of data into memory.

        Args:
            blob_id (uuid.UUID): The UUID of the blob.
            user_id (Optional[Union[str, uuid.UUID]], optional): ID of the user
                who's loading the blob.

        Returns:
            schema.NendoBlob: The loaded blob.
        """
        raise NotImplementedError

    @abstractmethod
    def remove_blob(
        self,
        blob_id: uuid.UUID,
        remove_resources: bool = True,
        user_id: Optional[uuid.UUID] = None,
    ) -> bool:
        """Deletes a blob of data.

        Args:
            blob_id (uuid.UUID): The UUID of the blob.
            remove_resources (bool): If True, remove associated resources.
            user_id (Optional[Union[str, uuid.UUID]], optional): ID of the user
                who's removing the blob.

        Returns:
            success (bool): True if removal was successful, False otherwise
        """
        raise NotImplementedError

    # ==================================
    #
    # MISCELLANEOUS MANAGEMENT FUNCTIONS
    #
    # ==================================

    def get_track_or_collection(
        self,
        target_id: Union[str, uuid.UUID],
    ) -> Union[NendoTrack, NendoCollection]:
        """Return a track or a collection based on the given target_id.

        Args:
            target_id (Union[str, uuid.UUID]): The target ID to obtain.

        Returns:
            Union[NendoTrack, NendoCollection]: The track or the collection.
        """
        target_id = ensure_uuid(target_id)
        collection = self.get_collection(target_id)
        if collection is not None:
            return collection

        # assume the id is a track id
        return self.get_track(target_id)

    def verify(self, action: Optional[str] = None, user_id: str = "") -> None:
        """Verify the library's integrity.

        Args:
            action (Optional[str], optional): Default action to choose when an
                inconsistency is detected. Choose between (i)gnore and (r)emove.
        """
        original_config = {}
        try:
            original_config["stream_mode"] = self.config.stream_mode
            original_config["stream_chunk_size"] = self.config.stream_chunk_size
            self.config.stream_mode = False
            self.config.stream_chunk_size = 16
            for track in self.get_tracks():
                if not self.storage_driver.file_exists(
                    file_name=track.resource.file_name,
                    user_id=user_id,
                ):
                    action = (
                        action
                        or input(
                            f"Inconsistency detected: {track.resource.src} "
                            "does not exist. Please choose an action:\n"
                            "(i) ignore - (r) remove",
                        ).lower()
                    )
                    if action == "i":
                        self.logger.warning(
                            "Detected missing file "
                            f"{track.resource.src} but instructed "
                            "to ignore.",
                        )
                        continue
                    if action == "r":
                        self.logger.info(
                            f"Removing track with ID {track.id} "
                            f"due to missing file {track.resource.src}",
                        )
                        self.remove_track(
                            track_id=track.id,
                            remove_plugin_data=True,
                            remove_relationships=True,
                            remove_resources=False,
                        )
            for library_file in self.storage_driver.list_files(user_id=user_id):
                file_without_ext = os.path.splitext(library_file)[0]
                if len(self.find_tracks(value=file_without_ext)) == 0:
                    action = (
                        action
                        or input(
                            f"Inconsistency detected: File {library_file} "
                            "cannot be fonud in database. Please choose an action:\n"
                            "(i) ignore - (r) remove",
                        ).lower()
                    )
                    if action == "i":
                        self.logger.warning(
                            f"Detected orphaned file {library_file} "
                            f"but instructed to ignore.",
                        )
                        continue
                    if action == "r":
                        self.logger.info(f"Removing orphaned file {library_file}")
                        self.storage_driver.remove_file(
                            file_name=library_file,
                            user_id=user_id,
                        )

        finally:
            self.config.stream_mode = original_config["stream_mode"]
            self.config.stream_chunk_size = original_config["stream_chunk_size"]

    @abstractmethod
    def reset(
        self,
        force: bool = False,
        user_id: Optional[Union[str, uuid.UUID]] = None,
    ) -> None:
        """Reset the nendo library.

        Erase all tracks, collections and relationships.
        Ask before erasing.

        Args:
            force (bool, optional): Flag that specifies whether to ask the user for
                confirmation of the operation. Default is to ask the user.
            user_id (Optional[Union[str, uuid.UUID]], optional): ID of the user
                who's resetting the library. If none is given, the configured
                nendo default user will be used.
        """
        raise NotImplementedError

    def __str__(self):
        output = f"{self.plugin_name}, version {self.plugin_version}:\n"
        output += f"{len(self)} tracks"
        return output

    @property
    def plugin_type(self) -> str:
        """Return type of plugin."""
        return "LibraryPlugin"



# -*- encoding: utf-8 -*-
"""Nendo core main class."""
from __future__ import annotations

import importlib
import inspect
import json
import logging
import os
import sys
from functools import lru_cache
from typing import List, Optional

from nendo import library as lib
from nendo import schema
from nendo.config import NendoConfig, get_settings

settings = get_settings()
nendo_logger = logging.getLogger("nendo")
log_fmt = "[%(asctime)s.%(msecs)03dZ] %(name)s         %(levelname)s %(message)s"


class Nendo:
    """Main class that runs nendo core."""

    _instance = None
    library: schema.NendoLibraryPlugin = None
    plugins: schema.NendoPluginRegistry = None
    logger: logging.Logger = None
    config: NendoConfig = None

    def __init__(
        self,
        config: Optional[NendoConfig] = None,
        plugins: Optional[List[str]] = None,
        logger: Optional[logging.Logger] = None,
    ) -> None:
        """Create a new nendo instance.

        Args:
            config (NendoConfig, optional): Custom settings to apply to this instance.
            plugins (str, optional): list of plugins to load. Defaults to None.
            logger (logging.Logger, optional): Logger to use. Defaults to none.
        """
        if hasattr(self, "_initialized") and self._initialized:
            return
        self.logger = logger or nendo_logger
        self.config = config or settings
        self.plugins = schema.NendoPluginRegistry()
        plugin_names = plugins or self.config.plugins
        if self.config.log_file_path == "":
            logging.basicConfig(
                level=self.config.log_level.upper(),
                datefmt="%Y-%m-%dT%H:%M:%S",
                format=log_fmt,
            )
        else:
            if (
                not os.path.isdir(os.path.dirname(self.config.log_file_path))
                or os.path.basename(self.config.log_file_path) == ""
            ):
                self.logger.error(
                    "Config var log_file_path has been set but path is not "
                    "valid or no filename specified. Please fix your configuration.",
                )
                sys.exit(-1)
            logging.basicConfig(
                filename=self.config.log_file_path,
                level=self.config.log_level.upper(),
                datefmt="%Y-%m-%dT%H:%M:%S",
                format=log_fmt,
            )
        self._load_plugins(plugin_names=plugin_names)
        # initialize nendo library
        if self.config.library_plugin == "default":
            self.logger.info("Using DuckDBLibrary")
            self.library = lib.DuckDBLibrary(
                nendo_instance=self,
                config=self.config,
                logger=self.logger,
                plugin_name="DuckDBLibrary",
                plugin_version="0.1.1",
            )
        else:
            self.logger.info(f"Loading {self.config.library_plugin}")
            self.library = self._load_plugin(module_name=self.config.library_plugin)
        self._initialized = True

    # this turns the nendo class effectively into a singleton
    def __new__(cls, *args, **kwargs):  # noqa: D102, ARG003
        if not cls._instance:
            cls._instance = super(Nendo, cls).__new__(cls)
            cls._instance._initialized = False
        return cls._instance

    # this makes all library functions callable on the nendo instance itself
    def __getattr__(self, attr):
        if not hasattr(self.library, attr):
            raise AttributeError(f"Function '{attr}' unknown.")
        return getattr(self.library, attr)

    def _load_plugin(self, module_name: str) -> schema.NendoPlugin:
        package_spec = importlib.util.find_spec(module_name)
        if package_spec is not None:
            try:
                module_class = importlib.import_module(module_name)
                for cl in inspect.getmembers(module_class, inspect.isclass):
                    # NOTE currently, we only support one functional class per plugin
                    if issubclass(cl[1], schema.NendoPlugin):
                        self.logger.debug(
                            "Adding Nendo plugin %s from %s",
                            cl,
                            module_name,
                        )
                        plugin_name = cl[1].__name__
                        plugin_version = importlib.metadata.version(module_name)
                        plugin_instance: schema.NendoPlugin = getattr(
                            module_class,
                            plugin_name,
                        )(
                            nendo_instance=self,
                            config=self.config,
                            logger=self.logger,
                            plugin_name=module_name,
                            plugin_version=plugin_version,
                        )
                        return plugin_instance
            except Exception as e:  # noqa: BLE001
                raise schema.NendoPluginLoadingError(
                    f"Failed to import plugin '{module_name}'. Error: {e}",
                ) from None
        else:
            raise schema.NendoPluginLoadingError(
                f"Plugin {module_name} not installed in system. "
                "Please use e.g. pip to install it.",
            )

    def _load_plugins(self, plugin_names: List[str] = settings.plugins) -> None:
        """Load the specified nendo plugins."""
        for module_name in plugin_names:
            plugin_instance = self._load_plugin(module_name=module_name)
            self.plugins.add(
                plugin_name=plugin_instance.plugin_name,
                version=plugin_instance.plugin_version,
                plugin_instance=plugin_instance,
            )

        plugins_str = str(self.plugins)
        for line in plugins_str.split("\n"):
            self.logger.info(line)

    def __str__(self):
        output = "Nendo\n"
        output += "Config: "
        output += json.dumps(json.loads(self.config.json()), indent=4)
        # output += f"Library: {self.library}\n"
        output += f"\n{self.plugins}"
        return output


@lru_cache()
def get_nendo():
    """Get the nendo instance singledton, cached."""
    return Nendo()
